# Unix SMB/CIFS implementation.
# Copyright (C) Martin Kraemer 2019 <mk.maddin@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
from samba.tests.samba_tool.base import SambaToolCmdTest
import re

class DSaclSetSddlTestCase(SambaToolCmdTest):
    """Tests for samba-tool dsacl set --sddl subcommand"""
    sddl       = "(OA;CIIO;RPWP;aaaaaaaa-1111-bbbb-2222-dddddddddddd;33333333-eeee-4444-ffff-555555555555;PS)"
    sddl_lc    = "(OA;CIIO;RPWP;aaaaaaaa-1111-bbbb-2222-dddddddddddd;33333333-eeee-4444-ffff-555555555555;PS)"
    sddl_uc    = "(OA;CIIO;RPWP;AAAAAAAA-1111-BBBB-2222-DDDDDDDDDDDD;33333333-EEEE-4444-FFFF-555555555555;PS)"
    sddl_sid   = "(OA;CIIO;RPWP;aaaaaaaa-1111-bbbb-2222-dddddddddddd;33333333-eeee-4444-ffff-555555555555;S-1-5-10)"
    sddl_multi = "(OA;CIIO;RPWP;aaaaaaaa-1111-bbbb-2222-dddddddddddd;33333333-eeee-4444-ffff-555555555555;PS)(OA;CIIO;RPWP;cccccccc-9999-ffff-8888-eeeeeeeeeeee;77777777-dddd-6666-bbbb-555555555555;PS)"

    def setUp(self):
        super().setUp()
        self.samdb = self.getSamDB("-H", "ldap://%s" % os.environ["DC_SERVER"],"-U%s%%%s" % (os.environ["DC_USERNAME"], os.environ["DC_PASSWORD"]))
        self.dn="OU=DSaclSetSddlTestCase,%s" % self.samdb.domain_dn()
        self.samdb.create_ou(self.dn)

    def tearDown(self):
        super().tearDown()
        # clean-up the created test ou
        self.samdb.delete(self.dn)

    def test_sddl(self):
        """Tests if a sddl string can be added 'the normal way'"""
        (result, out, err) = self.runsubcmd("dsacl", "set","--objectdn=%s" % self.dn, "--sddl=%s" % self.sddl)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        #extract only the two sddl strings from samba-tool output
        acl_list=re.findall('.*descriptor for.*:\n(.*?)\n',out)
        self.assertNotEqual(acl_list[0], acl_list[1], "new and old SDDL string differ")
        self.assertMatch(acl_list[1], self.sddl, "new SDDL string should be contained within second sddl output")

    def test_sddl_set_get(self):
        """Tests if a sddl string can be added 'the normal way' and the output of 'get' is the same"""
        (result, out, err) = self.runsubcmd("dsacl", "get",
                                            "--objectdn=%s" % self.dn)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        #extract only the two sddl strings from samba-tool output
        acl_list_get=re.findall('^descriptor for.*:\n(.*?)\n', out)

        (result, out, err) = self.runsubcmd("dsacl", "set",
                                            "--objectdn=%s" % self.dn,
                                            "--sddl=%s" % self.sddl)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        #extract only the two sddl strings from samba-tool output
        acl_list_old=re.findall('old descriptor for.*:\n(.*?)\n', out)
        self.assertEqual(acl_list_old, acl_list_get,
                         "output of dsacl get should be the same as before set")

        acl_list=re.findall('new descriptor for.*:\n(.*?)\n', out)

        (result, out, err) = self.runsubcmd("dsacl", "get",
                                            "--objectdn=%s" % self.dn)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        #extract only the two sddl strings from samba-tool output
        acl_list_get2=re.findall('^descriptor for.*:\n(.*?)\n', out)
        self.assertEqual(acl_list, acl_list_get2,
                         "output of dsacl get should be the same as after set")

    def test_multisddl(self):
        """Tests if we can add multiple, different sddl strings at the same time"""
        (result, out, err) = self.runsubcmd("dsacl", "set","--objectdn=%s" % self.dn, "--sddl=%s" % self.sddl_multi)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        #extract only the two sddl strings from samba-tool output
        acl_list=re.findall('.*descriptor for.*:\n(.*?)\n',out)
        for ace in re.findall(r'\(.*?\)',self.sddl_multi):
            self.assertMatch(acl_list[1], ace, "new SDDL string should be contained within second sddl output")

    def test_duplicatesddl(self):
        """Tests if an already existing sddl string can be added causing duplicate entry"""
        acl_list = self._double_sddl_check(self.sddl,self.sddl)
        self.assertEqual(acl_list[0],acl_list[1])

    def test_casesensitivesddl(self):
        """Tests if an already existing sddl string can be added in different cases causing duplicate entry"""
        acl_list = self._double_sddl_check(self.sddl_lc,self.sddl_uc)
        self.assertEqual(acl_list[0],acl_list[1])

    def test_sidsddl(self):
        """Tests if an already existing sddl string can be added with SID instead of SDDL SIDString causing duplicate entry"""
        acl_list = self._double_sddl_check(self.sddl,self.sddl_sid)
        self.assertEqual(acl_list[0],acl_list[1])

    def test_twosddl(self):
        """Tests if an already existing sddl string can be added by using it twice/in combination with non existing sddl string causing duplicate entry"""
        acl_list = self._double_sddl_check(self.sddl,self.sddl + self.sddl)
        self.assertEqual(acl_list[0],acl_list[1])

    def _double_sddl_check(self,sddl1,sddl2):
        """Adds two sddl strings and checks if there was an ace change after the second adding"""
        (result, out, err) = self.runsubcmd("dsacl", "set","--objectdn=%s" % self.dn, "--sddl=%s" % sddl1)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        acl_list = re.findall('.*descriptor for.*:\n(.*?)\n',out)
        self.assertMatch(acl_list[1], sddl1, "new SDDL string should be contained within second sddl output - is not")
        #add sddl2
        (result, out, err) = self.runsubcmd("dsacl", "set","--objectdn=%s" % self.dn, "--sddl=%s" % sddl2)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        acl_list = re.findall('.*descriptor for.*:\n(.*?)\n',out)
        return acl_list

    def test_add_delete_sddl(self):
        """Tests if a sddl string can be added 'the normal way', deleted and
        final state is the same as initial.
        """
        (result, out, err) = self.runsubcmd("dsacl", "get",
                                            "--objectdn=%s" % self.dn)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        # extract only the two sddl strings from samba-tool output
        acl_list_orig = re.findall('^descriptor for.*:\n(.*?)\n', out)[0]

        (result, out, err) = self.runsubcmd("dsacl", "set",
                                            "--objectdn=%s" % self.dn,
                                            "--sddl=%s" % self.sddl)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        acl_list_added = re.findall('new descriptor for.*:\n(.*?)\n', out)[0]
        self.assertNotEqual(acl_list_added, acl_list_orig, "After adding the SD should be different.")
        self.assertMatch(acl_list_added, self.sddl, "The added ACE should be part of the new SD.")

        (result, out, err) = self.runsubcmd("dsacl", "delete",
                                            "--objectdn=%s" % self.dn,
                                            "--sddl=%s" % self.sddl)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        acl_list_final = re.findall('new descriptor for.*:\n(.*?)\n', out)[0]
        self.assertEqual(acl_list_orig, acl_list_final,
                         "output of dsacl delete should be the same as before adding")

        (result, out, err) = self.runsubcmd("dsacl", "get",
                                            "--objectdn=%s" % self.dn)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        # extract only the two sddl strings from samba-tool output
        acl_list_final_get = re.findall('^descriptor for.*:\n(.*?)\n', out)[0]
        self.assertEqual(acl_list_orig, acl_list_final_get,
                         "output of dsacl get should be the same as after adding and deleting again")

    def test_delete(self):
        # add sddl_multi first
        (result, out, err) = self.runsubcmd("dsacl", "set",
                                            "--objectdn=%s" % self.dn,
                                            "--sddl=%s" % self.sddl_multi)

        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        # delete sddl
        (result, out, err) = self.runsubcmd("dsacl", "delete",
                                            "--objectdn=%s" % self.dn,
                                            "--sddl=%s" % self.sddl)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        acl_list_deleted = re.findall('new descriptor for.*:\n(.*?)\n', out)[0]

        self.assertNotRegex(acl_list_deleted, re.escape(self.sddl))
        left_sddl = self.sddl_multi.replace(self.sddl, "")
        self.assertRegex(acl_list_deleted, re.escape(left_sddl))

    def test_delete_twice(self):
        """Tests if deleting twice the same ACEs returns the expected warning."""
        # add sddl_multi first
        (result, out, err) = self.runsubcmd("dsacl", "set",
                                            "--objectdn=%s" % self.dn,
                                            "--sddl=%s" % self.sddl_multi)

        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")

        # delete sddl
        (result, out, err) = self.runsubcmd("dsacl", "delete",
                                            "--objectdn=%s" % self.dn,
                                            "--sddl=%s" % self.sddl)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")

        # delete sddl_multi
        (result, out, err) = self.runsubcmd("dsacl", "delete",
                                            "--objectdn=%s" % self.dn,
                                            "--sddl=%s" % self.sddl_multi)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        self.assertRegex(out, "WARNING", "Should throw a warning about deleting non existent ace.")
        warn = re.findall("WARNING: (.*?)\n", out)[0]
        left_sddl = self.sddl_multi.replace(self.sddl, "")
        self.assertRegex(warn, re.escape(self.sddl), "Should point out the non existent ace.")
        self.assertNotRegex(warn, re.escape(left_sddl),
                            "Should not complain about all aces, since one of them is not deleted twice.")



# Samba-specific bits for optparse
# Copyright (C) Jelmer Vernooij <jelmer@samba.org> 2007
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Support for parsing Samba-related command-line options."""

__docformat__ = "restructuredText"

import optparse
import os
import sys
from abc import ABCMeta, abstractmethod
from copy import copy

from samba.credentials import (
    Credentials,
    AUTO_USE_KERBEROS,
    DONT_USE_KERBEROS,
    MUST_USE_KERBEROS,
)
from samba._glue import get_burnt_commandline


def check_bytes(option, opt, value):
    """Custom option type to allow the input of sizes using byte, kb, mb ...

    units, e.g. 2Gb, 4KiB ...
       e.g. Option("--size", type="bytes", metavar="SIZE")
    """

    multipliers = {"B": 1,
                   "KB": 1024,
                   "MB": 1024 * 1024,
                   "GB": 1024 * 1024 * 1024}

    # strip out any spaces
    v = value.replace(" ", "")

    # extract the numeric prefix
    digits = ""
    while v and v[0:1].isdigit() or v[0:1] == '.':
        digits += v[0]
        v = v[1:]

    try:
        m = float(digits)
    except ValueError:
        msg = ("{0} option requires a numeric value, "
               "with an optional unit suffix").format(opt)
        raise optparse.OptionValueError(msg)

    # strip out the 'i' and convert to upper case so
    # kib Kib kb KB are all equivalent
    suffix = v.upper().replace("I", "")
    try:
        return m * multipliers[suffix]
    except KeyError as k:
        msg = ("{0} invalid suffix '{1}', "
               "should be B, Kb, Mb or Gb").format(opt, v)
        raise optparse.OptionValueError(msg)


class OptionMissingError(optparse.OptionValueError):
    """One or more Options with required=True is missing."""

    def __init__(self, options):
        """Raised when required Options are missing from the command line.

        :param options: list of 1 or more option
        """
        self.options = options

    def __str__(self):
        if len(self.options) == 1:
            missing = self.options[0]
            return f"Argument {missing} is required."
        else:
            options = sorted([str(option) for option in self.options])
            missing = ", ".join(options)
            return f"The arguments {missing} are required."


class ValidationError(Exception):
    """ValidationError is the exception raised by validators.

    Should be raised from the __call__ method of the Validator subclass.
    """
    pass


class Validator(metaclass=ABCMeta):
    """Base class for Validators used by SambaOption.

    Subclass this to make custom validators and implement __call__.
    """

    @abstractmethod
    def __call__(self, field, value):
        pass


class Option(optparse.Option):
    ATTRS = optparse.Option.ATTRS + ["required", "validators"]
    TYPES = optparse.Option.TYPES + ("bytes",)
    TYPE_CHECKER = copy(optparse.Option.TYPE_CHECKER)
    TYPE_CHECKER["bytes"] = check_bytes

    def run_validators(self, opt, value):
        """Runs the list of validators on the current option."""
        validators = getattr(self, "validators") or []
        for validator in validators:
            validator(opt, value)

    def convert_value(self, opt, value):
        """Override convert_value to run validators just after.

        This can also be done in process() but there we would have to
        replace the entire method.
        """
        value = super().convert_value(opt, value)
        self.run_validators(opt, value)
        return value


class OptionParser(optparse.OptionParser):
    """Samba OptionParser, adding support for required=True on Options."""

    def __init__(self,
                 usage=None,
                 option_list=None,
                 option_class=Option,
                 version=None,
                 conflict_handler="error",
                 description=None,
                 formatter=None,
                 add_help_option=True,
                 prog=None,
                 epilog=None):
        """
        Ensure that option_class defaults to the Samba one.
        """
        super().__init__(usage, option_list, option_class, version,
                         conflict_handler, description, formatter,
                         add_help_option, prog, epilog)

    def check_values(self, values, args):
        """Loop through required options if value is missing raise exception."""
        missing = []
        for option in self._get_all_options():
            if option.required:
                value = getattr(values, option.dest)
                if value is None:
                    missing.append(option)

        if missing:
            raise OptionMissingError(missing)

        return super().check_values(values, args)


class OptionGroup(optparse.OptionGroup):
    """Samba OptionGroup base class.

    Provides a generic set_option method to be used as Option callback,
    so that one doesn't need to be created for every available Option.

    Also overrides the add_option method, so it correctly initialises
    the defaults on the OptionGroup.
    """

    def add_option(self, *args, **kwargs):
        """Override add_option so it applies defaults during constructor."""
        opt = super().add_option(*args, **kwargs)
        default = None if opt.default == optparse.NO_DEFAULT else opt.default
        self.set_option(opt, opt.get_opt_string(), default, self.parser)
        return opt

    def set_option(self, option, opt_str, arg, parser):
        """Callback to set the attribute based on the Option dest name."""
        dest = option.dest or option._long_opts[0][2:].replace("-", "_")
        setattr(self, dest, arg)


class SambaOptions(OptionGroup):
    """General Samba-related command line options."""

    def __init__(self, parser):
        from samba import fault_setup
        fault_setup()

        # This removes passwords from the commandline via
        # setproctitle() but makes no change to python sys.argv so we
        # can continue to process as normal
        #
        # get_burnt_commandline returns None if no change is needed
        new_proctitle = get_burnt_commandline(sys.argv)
        if new_proctitle is not None:
            try:
                import setproctitle
                setproctitle.setproctitle(new_proctitle)

            except ModuleNotFoundError:
                msg = ("WARNING: Using passwords on command line is insecure. "
                       "Installing the setproctitle python module will hide "
                       "these from shortly after program start.\n")
                sys.stderr.write(msg)
                sys.stderr.flush()

        from samba.param import LoadParm
        super().__init__(parser, "Samba Common Options")
        self.add_option("-s", "--configfile", action="callback",
                        type=str, metavar="FILE", help="Configuration file",
                        callback=self._load_configfile)
        self.add_option("-d", "--debuglevel", action="callback",
                        type=str, metavar="DEBUGLEVEL", help="debug level",
                        callback=self._set_debuglevel)
        self.add_option("--option", action="callback",
                        type=str, metavar="OPTION",
                        help="set smb.conf option from command line",
                        callback=self._set_option)
        self.add_option("--realm", action="callback",
                        type=str, metavar="REALM", help="set the realm name",
                        callback=self._set_realm)
        self._configfile = None
        self._lp = LoadParm()
        self.realm = None

    def get_loadparm_path(self):
        """Return path to the smb.conf file specified on the command line."""
        return self._configfile

    def _load_configfile(self, option, opt_str, arg, parser):
        self._configfile = arg

    def _set_debuglevel(self, option, opt_str, arg, parser):
        try:
            self._lp.set('debug level', arg)
        except RuntimeError:
            raise optparse.OptionValueError(
                f"invalid -d/--debug value: '{arg}'")
        parser.values.debuglevel = arg

    def _set_realm(self, option, opt_str, arg, parser):
        try:
            self._lp.set('realm', arg)
        except RuntimeError:
            raise optparse.OptionValueError(
                f"invalid --realm value: '{arg}'")
        self.realm = arg

    def _set_option(self, option, opt_str, arg, parser):
        if arg.find('=') == -1:
            raise optparse.OptionValueError(
                "--option option takes a 'a=b' argument")
        a = arg.split('=', 1)
        try:
            self._lp.set(a[0], a[1])
        except Exception as e:
            raise optparse.OptionValueError(
                "invalid --option option value %r: %s" % (arg, e))

    def get_loadparm(self):
        """Return loadparm object with data specified on the command line."""
        if self._configfile is not None:
            self._lp.load(self._configfile)
        elif os.getenv("SMB_CONF_PATH") is not None:
            self._lp.load(os.getenv("SMB_CONF_PATH"))
        else:
            self._lp.load_default()
        return self._lp


class Samba3Options(SambaOptions):
    """General Samba-related command line options with an s3 param."""

    def __init__(self, parser):
        super().__init__(parser)
        from samba.samba3 import param as s3param
        self._lp = s3param.get_context()


class HostOptions(OptionGroup):
    """Command line options for connecting to target host or database."""

    def __init__(self, parser):
        super().__init__(parser, "Host Options")

        self.add_option("-H", "--URL",
                        help="LDB URL for database or target server",
                        type=str, metavar="URL", action="callback",
                        callback=self.set_option, dest="H")


class VersionOptions(OptionGroup):
    """Command line option for printing Samba version."""
    def __init__(self, parser):
        super().__init__(parser, "Version Options")
        self.add_option("-V", "--version", action="callback",
                        callback=self._display_version,
                        help="Display version number")

    def _display_version(self, option, opt_str, arg, parser):
        import samba
        print(samba.version)
        sys.exit(0)


def parse_kerberos_arg_legacy(arg, opt_str):
    if arg.lower() in ["yes", 'true', '1']:
        return MUST_USE_KERBEROS
    elif arg.lower() in ["no", 'false', '0']:
        return DONT_USE_KERBEROS
    elif arg.lower() in ["auto"]:
        return AUTO_USE_KERBEROS
    else:
        raise optparse.OptionValueError("invalid %s option value: %s" %
                                        (opt_str, arg))


def parse_kerberos_arg(arg, opt_str):
    if arg.lower() == 'required':
        return MUST_USE_KERBEROS
    elif arg.lower() == 'desired':
        return AUTO_USE_KERBEROS
    elif arg.lower() == 'off':
        return DONT_USE_KERBEROS
    else:
        raise optparse.OptionValueError("invalid %s option value: %s" %
                                        (opt_str, arg))


class CredentialsOptions(OptionGroup):
    """Command line options for specifying credentials."""

    def __init__(self, parser, special_name=None):
        self.special_name = special_name
        if special_name is not None:
            self.section = "Credentials Options (%s)" % special_name
        else:
            self.section = "Credentials Options"

        self.ask_for_password = True
        self.ipaddress = None
        self.machine_pass = False
        super().__init__(parser, self.section)
        self._add_option("--simple-bind-dn", metavar="DN", action="callback",
                         callback=self._set_simple_bind_dn, type=str,
                         help="DN to use for a simple bind")
        self._add_option("--password", metavar="PASSWORD", action="callback",
                         help="Password", type=str, callback=self._set_password)
        self._add_option("-U", "--username", metavar="USERNAME",
                         action="callback", type=str,
                         help="Username", callback=self._parse_username)
        self._add_option("-W", "--workgroup", metavar="WORKGROUP",
                         action="callback", type=str,
                         help="Workgroup", callback=self._parse_workgroup)
        self._add_option("-N", "--no-pass", action="callback",
                         help="Don't ask for a password",
                         callback=self._set_no_password)
        self._add_option("", "--ipaddress", metavar="IPADDRESS",
                         action="callback", type=str,
                         help="IP address of server",
                         callback=self._set_ipaddress)
        self._add_option("-P", "--machine-pass",
                         action="callback",
                         help="Use stored machine account password",
                         callback=self._set_machine_pass)
        self._add_option("--use-kerberos", metavar="desired|required|off",
                         action="callback", type=str,
                         help="Use Kerberos authentication", callback=self._set_kerberos)
        self._add_option("--use-krb5-ccache", metavar="KRB5CCNAME",
                         action="callback", type=str,
                         help="Kerberos Credentials cache",
                         callback=self._set_krb5_ccache)
        self._add_option("-A", "--authentication-file", metavar="AUTHFILE",
                         action="callback", type=str,
                         help="Authentication file",
                         callback=self._set_auth_file)

        # LEGACY
        self._add_option("-k", "--kerberos", metavar="KERBEROS",
                         action="callback", type=str,
                         help="DEPRECATED: Migrate to --use-kerberos", callback=self._set_kerberos_legacy)
        self.creds = Credentials()

    def _add_option(self, *args1, **kwargs):
        if self.special_name is None:
            return self.add_option(*args1, **kwargs)

        args2 = ()
        for a in args1:
            if not a.startswith("--"):
                continue
            args2 += (a.replace("--", "--%s-" % self.special_name),)
        self.add_option(*args2, **kwargs)

    def _parse_username(self, option, opt_str, arg, parser):
        self.creds.parse_string(arg)
        self.machine_pass = False

    def _parse_workgroup(self, option, opt_str, arg, parser):
        self.creds.set_domain(arg)

    def _set_password(self, option, opt_str, arg, parser):
        self.creds.set_password(arg)
        self.ask_for_password = False
        self.machine_pass = False

    def _set_no_password(self, option, opt_str, arg, parser):
        self.ask_for_password = False

    def _set_machine_pass(self, option, opt_str, arg, parser):
        self.machine_pass = True

    def _set_ipaddress(self, option, opt_str, arg, parser):
        self.ipaddress = arg

    def _set_kerberos_legacy(self, option, opt_str, arg, parser):
        print('WARNING: The option -k|--kerberos is deprecated!')
        self.creds.set_kerberos_state(parse_kerberos_arg_legacy(arg, opt_str))

    def _set_kerberos(self, option, opt_str, arg, parser):
        self.creds.set_kerberos_state(parse_kerberos_arg(arg, opt_str))

    def _set_simple_bind_dn(self, option, opt_str, arg, parser):
        self.creds.set_bind_dn(arg)

    def _set_krb5_ccache(self, option, opt_str, arg, parser):
        self.creds.set_kerberos_state(MUST_USE_KERBEROS)
        self.creds.set_named_ccache(arg)

    def _set_auth_file(self, option, opt_str, arg, parser):
        if os.path.exists(arg):
            self.creds.parse_file(arg)
            self.ask_for_password = False
            self.machine_pass = False

    def get_credentials(self, lp, fallback_machine=False):
        """Obtain the credentials set on the command-line.

        :param lp: Loadparm object to use.
        :return: Credentials object
        """
        self.creds.guess(lp)
        if self.machine_pass:
            self.creds.set_machine_account(lp)
        elif self.ask_for_password:
            self.creds.set_cmdline_callbacks()

        # possibly fallback to using the machine account, if we have
        # access to the secrets db
        if fallback_machine and not self.creds.authentication_requested():
            try:
                self.creds.set_machine_account(lp)
            except Exception:
                pass

        return self.creds


class CredentialsOptionsDouble(CredentialsOptions):
    """Command line options for specifying credentials of two servers."""

    def __init__(self, parser):
        super().__init__(parser)
        self.no_pass2 = True
        self.add_option("--simple-bind-dn2", metavar="DN2", action="callback",
                        callback=self._set_simple_bind_dn2, type=str,
                        help="DN to use for a simple bind")
        self.add_option("--password2", metavar="PASSWORD2", action="callback",
                        help="Password", type=str,
                        callback=self._set_password2)
        self.add_option("--username2", metavar="USERNAME2",
                        action="callback", type=str,
                        help="Username for second server",
                        callback=self._parse_username2)
        self.add_option("--workgroup2", metavar="WORKGROUP2",
                        action="callback", type=str,
                        help="Workgroup for second server",
                        callback=self._parse_workgroup2)
        self.add_option("--no-pass2", action="store_true",
                        help="Don't ask for a password for the second server")
        self.add_option("--use-kerberos2", metavar="desired|required|off",
                        action="callback", type=str,
                        help="Use Kerberos authentication", callback=self._set_kerberos2)

        # LEGACY
        self.add_option("--kerberos2", metavar="KERBEROS2",
                        action="callback", type=str,
                        help="Use Kerberos", callback=self._set_kerberos2_legacy)
        self.creds2 = Credentials()

    def _parse_username2(self, option, opt_str, arg, parser):
        self.creds2.parse_string(arg)

    def _parse_workgroup2(self, option, opt_str, arg, parser):
        self.creds2.set_domain(arg)

    def _set_password2(self, option, opt_str, arg, parser):
        self.creds2.set_password(arg)
        self.no_pass2 = False

    def _set_kerberos2_legacy(self, option, opt_str, arg, parser):
        self.creds2.set_kerberos_state(parse_kerberos_arg(arg, opt_str))

    def _set_kerberos2(self, option, opt_str, arg, parser):
        self.creds2.set_kerberos_state(parse_kerberos_arg(arg, opt_str))

    def _set_simple_bind_dn2(self, option, opt_str, arg, parser):
        self.creds2.set_bind_dn(arg)

    def get_credentials2(self, lp, guess=True):
        """Obtain the credentials set on the command-line.

        :param lp: Loadparm object to use.
        :param guess: Try guess Credentials from environment
        :return: Credentials object
        """
        if guess:
            self.creds2.guess(lp)
        elif not self.creds2.get_username():
            self.creds2.set_anonymous()

        if self.no_pass2:
            self.creds2.set_cmdline_callbacks()
        return self.creds2



# Unix SMB/CIFS implementation.
#
# authentication silos - authentication silo management
#
# Copyright (C) Catalyst.Net Ltd. 2023
#
# Written by Rob van der Linde <rob@catalyst.net.nz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import samba.getopt as options
from samba.netcmd import Command, CommandError, Option, SuperCommand
from samba.netcmd.domain.models import AuthenticationPolicy, AuthenticationSilo
from samba.netcmd.domain.models.exceptions import ModelError

from .silo_member import cmd_domain_auth_silo_member


class cmd_domain_auth_silo_list(Command):
    """List authentication silos on the domain."""

    synopsis = "%prog -H <URL> [options]"

    takes_optiongroups = {
        "sambaopts": options.SambaOptions,
        "credopts": options.CredentialsOptions,
        "hostopts": options.HostOptions,
    }

    takes_options = [
        Option("--json", help="Output results in JSON format.",
               dest="output_format", action="store_const", const="json"),
    ]

    def run(self, hostopts=None, sambaopts=None, credopts=None,
            output_format=None):

        ldb = self.ldb_connect(hostopts, sambaopts, credopts)

        # Authentication silos grouped by cn.
        try:
            silos = {silo.cn: silo.as_dict()
                     for silo in AuthenticationSilo.query(ldb)}
        except ModelError as e:
            raise CommandError(e)

        # Using json output format gives more detail.
        if output_format == "json":
            self.print_json(silos)
        else:
            for silo in silos.keys():
                self.outf.write(f"{silo}\n")


class cmd_domain_auth_silo_view(Command):
    """View an authentication silo on the domain."""

    synopsis = "%prog -H <URL> [options]"

    takes_optiongroups = {
        "sambaopts": options.SambaOptions,
        "credopts": options.CredentialsOptions,
        "hostopts": options.HostOptions,
    }

    takes_options = [
        Option("--name",
               help="Name of authentication silo to view (required).",
               dest="name", action="store", type=str, required=True),
    ]

    def run(self, hostopts=None, sambaopts=None, credopts=None, name=None):

        ldb = self.ldb_connect(hostopts, sambaopts, credopts)

        try:
            silo = AuthenticationSilo.get(ldb, cn=name)
        except ModelError as e:
            raise CommandError(e)

        # Check if silo exists first.
        if silo is None:
            raise CommandError(f"Authentication silo {name} not found.")

        # Display silo as JSON.
        self.print_json(silo.as_dict())


class cmd_domain_auth_silo_create(Command):
    """Create a new authentication silo on the domain."""

    synopsis = "%prog -H <URL> [options]"

    takes_optiongroups = {
        "sambaopts": options.SambaOptions,
        "credopts": options.CredentialsOptions,
        "hostopts": options.HostOptions,
    }

    takes_options = [
        Option("--name", help="Name of authentication silo (required).",
               dest="name", action="store", type=str, required=True),
        Option("--description",
               help="Optional description for authentication silo.",
               dest="description", action="store", type=str),
        Option("--user-authentication-policy",
               help="User account authentication policy.",
               dest="user_authentication_policy", action="store", type=str,
               metavar="USER_POLICY"),
        Option("--service-authentication-policy",
               help="Managed service account authentication policy.",
               dest="service_authentication_policy", action="store", type=str,
               metavar="SERVICE_POLICY"),
        Option("--computer-authentication-policy",
               help="Computer authentication policy.",
               dest="computer_authentication_policy", action="store", type=str,
               metavar="COMPUTER_POLICY"),
        Option("--protect",
               help="Protect authentication silo from accidental deletion.",
               dest="protect", action="store_true"),
        Option("--unprotect",
               help="Unprotect authentication silo from accidental deletion.",
               dest="unprotect", action="store_true"),
        Option("--audit",
               help="Only audit silo policies.",
               dest="audit", action="store_true"),
        Option("--enforce",
               help="Enforce silo policies.",
               dest="enforce", action="store_true")
    ]

    @staticmethod
    def get_policy(ldb, name):
        """Helper function to fetch auth policy or raise CommandError.

        :param ldb: Ldb connection
        :param name: Either the DN or name of authentication policy
        """
        try:
            return AuthenticationPolicy.lookup(ldb, name)
        except (LookupError, ValueError) as e:
            raise CommandError(e)

    def run(self, hostopts=None, sambaopts=None, credopts=None,
            name=None, description=None,
            user_authentication_policy=None,
            service_authentication_policy=None,
            computer_authentication_policy=None,
            protect=None, unprotect=None,
            audit=None, enforce=None):

        if protect and unprotect:
            raise CommandError("--protect and --unprotect cannot be used together.")
        if audit and enforce:
            raise CommandError("--audit and --enforce cannot be used together.")

        ldb = self.ldb_connect(hostopts, sambaopts, credopts)

        try:
            silo = AuthenticationSilo.get(ldb, cn=name)
        except ModelError as e:
            raise CommandError(e)

        # Make sure silo doesn't already exist.
        if silo is not None:
            raise CommandError(f"Authentication silo {name} already exists.")

        # New silo object.
        silo = AuthenticationSilo(cn=name, description=description)

        # Set user policy
        if user_authentication_policy:
            silo.user_authentication_policy = \
                self.get_policy(ldb, user_authentication_policy).dn

        # Set service policy
        if service_authentication_policy:
            silo.service_authentication_policy = \
                self.get_policy(ldb, service_authentication_policy).dn

        # Set computer policy
        if computer_authentication_policy:
            silo.computer_authentication_policy = \
                self.get_policy(ldb, computer_authentication_policy).dn

        # Either --enforce will be set or --audit but never both.
        # The default if both are missing is enforce=True.
        if enforce is not None:
            silo.enforced = enforce
        else:
            silo.enforced = not audit

        # Create silo
        try:
            silo.save(ldb)

            if protect:
                silo.protect(ldb)
        except ModelError as e:
            raise CommandError(e)

        # Authentication silo created successfully.
        self.outf.write(f"Created authentication silo: {name}\n")


class cmd_domain_auth_silo_modify(Command):
    """Modify an authentication silo on the domain."""

    synopsis = "%prog -H <URL> [options]"

    takes_optiongroups = {
        "sambaopts": options.SambaOptions,
        "credopts": options.CredentialsOptions,
        "hostopts": options.HostOptions,
    }

    takes_options = [
        Option("--name", help="Name of authentication silo (required).",
               dest="name", action="store", type=str, required=True),
        Option("--description",
               help="Optional description for authentication silo.",
               dest="description", action="store", type=str),
        Option("--user-authentication-policy",
               help="User account authentication policy.",
               dest="user_authentication_policy", action="store", type=str,
               metavar="USER_POLICY"),
        Option("--service-authentication-policy",
               help="Managed service account authentication policy.",
               dest="service_authentication_policy", action="store", type=str,
               metavar="SERVICE_POLICY"),
        Option("--computer-authentication-policy",
               help="Computer authentication policy.",
               dest="computer_authentication_policy", action="store", type=str,
               metavar="COMPUTER_POLICY"),
        Option("--protect",
               help="Protect authentication silo from accidental deletion.",
               dest="protect", action="store_true"),
        Option("--unprotect",
               help="Unprotect authentication silo from accidental deletion.",
               dest="unprotect", action="store_true"),
        Option("--audit",
               help="Only audit silo policies.",
               dest="audit", action="store_true"),
        Option("--enforce",
               help="Enforce silo policies.",
               dest="enforce", action="store_true")
    ]

    @staticmethod
    def get_policy(ldb, name):
        """Helper function to fetch auth policy or raise CommandError.

        :param ldb: Ldb connection
        :param name: Either the DN or name of authentication policy
        """
        try:
            return AuthenticationPolicy.lookup(ldb, name)
        except (LookupError, ModelError, ValueError) as e:
            raise CommandError(e)

    def run(self, hostopts=None, sambaopts=None, credopts=None,
            name=None, description=None,
            user_authentication_policy=None,
            service_authentication_policy=None,
            computer_authentication_policy=None,
            protect=None, unprotect=None,
            audit=None, enforce=None):

        if audit and enforce:
            raise CommandError("--audit and --enforce cannot be used together.")
        if protect and unprotect:
            raise CommandError("--protect and --unprotect cannot be used together.")

        ldb = self.ldb_connect(hostopts, sambaopts, credopts)

        try:
            silo = AuthenticationSilo.get(ldb, cn=name)
        except ModelError as e:
            raise CommandError(e)

        # Check if silo exists first.
        if silo is None:
            raise CommandError(f"Authentication silo {name} not found.")

        # Either --enforce will be set or --audit but never both.
        if enforce:
            silo.enforced = True
        elif audit:
            silo.enforced = False

        # Update the description.
        if description is not None:
            silo.description = description

        # Set or unset user policy.
        if user_authentication_policy == "":
            silo.user_authentication_policy = None
        elif user_authentication_policy:
            silo.user_authentication_policy = \
                self.get_policy(ldb, user_authentication_policy).dn

        # Set or unset service policy.
        if service_authentication_policy == "":
            silo.service_authentication_policy = None
        elif service_authentication_policy:
            silo.service_authentication_policy = \
                self.get_policy(ldb, service_authentication_policy).dn

        # Set or unset computer policy.
        if computer_authentication_policy == "":
            silo.computer_authentication_policy = None
        elif computer_authentication_policy:
            silo.computer_authentication_policy = \
                self.get_policy(ldb, computer_authentication_policy).dn

        # Update silo
        try:
            silo.save(ldb)

            if protect:
                silo.protect(ldb)
            elif unprotect:
                silo.unprotect(ldb)
        except ModelError as e:
            raise CommandError(e)

        # Silo updated successfully.
        self.outf.write(f"Updated authentication silo: {name}\n")


class cmd_domain_auth_silo_delete(Command):
    """Delete an authentication silo on the domain."""

    synopsis = "%prog -H <URL> [options]"

    takes_optiongroups = {
        "sambaopts": options.SambaOptions,
        "credopts": options.CredentialsOptions,
        "hostopts": options.HostOptions,
    }

    takes_options = [
        Option("--name", help="Name of authentication silo (required).",
               dest="name", action="store", type=str, required=True),
        Option("--force", help="Force delete protected authentication silo.",
               dest="force", action="store_true")
    ]

    def run(self, hostopts=None, sambaopts=None, credopts=None, name=None,
            force=None):

        ldb = self.ldb_connect(hostopts, sambaopts, credopts)

        try:
            silo = AuthenticationSilo.get(ldb, cn=name)
        except ModelError as e:
            raise CommandError(e)

        # Check if silo exists first.
        if silo is None:
            raise CommandError(f"Authentication silo {name} not found.")

        # Delete silo
        try:
            if force:
                silo.unprotect(ldb)

            silo.delete(ldb)
        except ModelError as e:
            if not force:
                raise CommandError(
                    f"{e}\nTry --force to delete protected authentication silos.")
            else:
                raise CommandError(e)

        # Authentication silo deleted successfully.
        self.outf.write(f"Deleted authentication silo: {name}\n")


class cmd_domain_auth_silo(SuperCommand):
    """Manage authentication silos on the domain."""

    subcommands = {
        "list": cmd_domain_auth_silo_list(),
        "view": cmd_domain_auth_silo_view(),
        "create": cmd_domain_auth_silo_create(),
        "modify": cmd_domain_auth_silo_modify(),
        "delete": cmd_domain_auth_silo_delete(),
        "member": cmd_domain_auth_silo_member(),
    }



# Unix SMB/CIFS implementation.
# Copyright Volker Lendecke <vl@samba.org> 2012
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Tests for samba.samba3.libsmb."""

from samba.samba3 import libsmb_samba_internal as libsmb
from samba.dcerpc import security
from samba.samba3 import param as s3param
from samba import credentials
from samba import (ntstatus,NTSTATUSError)
import samba.tests
import os

class LibsmbTests(samba.tests.TestCase):

    def setUp(self):
        self.lp = s3param.get_context()
        self.lp.load(samba.tests.env_get_var_value("SMB_CONF_PATH"))

        self.creds = credentials.Credentials()
        self.creds.guess(self.lp)
        self.creds.set_domain(samba.tests.env_get_var_value("DOMAIN"))
        self.creds.set_username(samba.tests.env_get_var_value("USERNAME"))
        self.creds.set_password(samba.tests.env_get_var_value("PASSWORD"))

        # Build the global inject file path
        server_conf = samba.tests.env_get_var_value("SERVERCONFFILE")
        server_conf_dir = os.path.dirname(server_conf)
        self.global_inject = os.path.join(server_conf_dir, "global_inject.conf")

        self.server_ip = samba.tests.env_get_var_value("SERVER_IP")

    def clean_file(self, conn, filename):
        try:
            conn.unlink(filename)
        except NTSTATUSError as e:
            if e.args[0] == ntstatus.NT_STATUS_FILE_IS_A_DIRECTORY:
                conn.rmdir(filename)
            elif not (e.args[0] == ntstatus.NT_STATUS_OBJECT_NAME_NOT_FOUND or
                      e.args[0] == ntstatus.NT_STATUS_OBJECT_PATH_NOT_FOUND):
                raise



# Unix SMB/CIFS implementation. Tests for kcc.graph_utils routines
# Copyright (C) Andrew Bartlett 2015
#
# Written by Douglas Bagnall <douglas.bagnall@catalyst.net.nz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Tests for samba.kcc.graph_utils"""

import samba
import samba.tests
from samba.kcc.graph_utils import GraphError
from samba.kcc.graph_utils import (verify_graph_complete,
                                   verify_graph_connected,
                                   verify_graph_connected_under_edge_failures,
                                   verify_graph_forest,
                                   verify_graph_connected_under_vertex_failures,
                                   verify_graph_no_lonely_vertices)

import itertools


def make_tree(vertices):
    if len(vertices) < 2:
        return ()
    remaining = set(vertices)
    used = set()
    edges = set()
    used.add(remaining.pop())
    used.add(remaining.pop())
    edges.add(tuple(used))
    while remaining:
        v = remaining.pop()
        w = used.pop()
        e = (w, v)
        edges.add(e)
        used.update(e)
    return tuple(edges)

# TODO: test directed graphs


class UndirectedGraphTests(samba.tests.TestCase):

    def setUp(self):
        super().setUp()
        vertices = tuple('abcdefgh')
        vertices2 = tuple('ijk')
        edges = tuple(itertools.combinations(vertices, 2))
        edges2 = tuple(itertools.combinations(vertices2, 2))
        line_edges = list(zip(vertices[1:], vertices[:-1]))
        ring_edges = line_edges + [(vertices[0], vertices[-1])]

        tree = make_tree(vertices)
        tree2 = make_tree(vertices2)

        self.complete_graph = [edges, vertices, vertices]

        self.disconnected_clusters = [edges + edges2,
                                      vertices + vertices2,
                                      vertices + vertices2]

        self.graph_with_unreachables = [edges,
                                        vertices + vertices2,
                                        vertices]

        self.ring = [ring_edges, vertices, vertices]
        self.line = [line_edges, vertices, vertices]

        self.tree = [tree, vertices, vertices]
        self.forest = [tree + tree2,
                       vertices + vertices2,
                       vertices + vertices2]

        self.unconnected_graph = ((), vertices, ())

    def assertGraphError(self, fn, *args):
        return self.assertRaises(GraphError, fn, *args)

    def test_graph_complete(self):
        fn = verify_graph_complete

        self.assertGraphError(fn, *self.disconnected_clusters)
        self.assertGraphError(fn, *self.graph_with_unreachables)
        self.assertGraphError(fn, *self.ring)
        self.assertGraphError(fn, *self.tree)

        self.assertIsNone(fn(*self.complete_graph))

    def test_graph_connected(self):
        fn = verify_graph_connected

        self.assertGraphError(fn, *self.disconnected_clusters)
        self.assertGraphError(fn, *self.graph_with_unreachables)
        self.assertGraphError(fn, *self.forest)
        self.assertGraphError(fn, *self.unconnected_graph)

        self.assertIsNone(fn(*self.line))
        self.assertIsNone(fn(*self.ring))
        self.assertIsNone(fn(*self.complete_graph))
        self.assertIsNone(fn(*self.tree))

    def test_graph_forest(self):
        fn = verify_graph_forest

        self.assertGraphError(fn, *self.disconnected_clusters)
        self.assertGraphError(fn, *self.graph_with_unreachables)
        self.assertGraphError(fn, *self.ring)

        self.assertIsNone(fn(*self.line))
        self.assertIsNone(fn(*self.tree))
        self.assertIsNone(fn(*self.forest))
        self.assertIsNone(fn(*self.unconnected_graph))

    def test_graph_connected_under_edge_failures(self):
        fn = verify_graph_connected_under_edge_failures

        self.assertGraphError(fn, *self.line)
        self.assertGraphError(fn, *self.tree)
        self.assertGraphError(fn, *self.forest)
        self.assertGraphError(fn, *self.disconnected_clusters)

        self.assertIsNone(fn(*self.ring))
        self.assertIsNone(fn(*self.complete_graph))

    def test_graph_connected_under_vertex_failures(self):
        # XXX no tests to distinguish this from the edge_failures case
        fn = verify_graph_connected_under_vertex_failures

        self.assertGraphError(fn, *self.line)
        self.assertGraphError(fn, *self.tree)
        self.assertGraphError(fn, *self.forest)
        self.assertGraphError(fn, *self.disconnected_clusters)

        self.assertIsNone(fn(*self.ring))
        self.assertIsNone(fn(*self.complete_graph))

    def test_graph_multi_edge_forest(self):
        pass

    def test_graph_no_lonely_vertices(self):
        fn = verify_graph_no_lonely_vertices
        self.assertGraphError(fn, *self.unconnected_graph)
        self.assertGraphError(fn, *self.graph_with_unreachables)

        self.assertIsNone(fn(*self.ring))
        self.assertIsNone(fn(*self.complete_graph))
        self.assertIsNone(fn(*self.line))
        self.assertIsNone(fn(*self.tree))
        self.assertIsNone(fn(*self.forest))

    def test_graph_no_unknown_vertices(self):
        pass



# Unix SMB/CIFS implementation.
#
# claim value type management
#
# Copyright (C) Catalyst.Net Ltd. 2023
#
# Written by Rob van der Linde <rob@catalyst.net.nz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import samba.getopt as options
from samba.netcmd import Command, CommandError, Option, SuperCommand
from samba.netcmd.domain.models import ValueType
from samba.netcmd.domain.models.exceptions import ModelError


class cmd_domain_claim_value_type_list(Command):
    """List claim values types on the domain."""

    synopsis = "%prog -H <URL> [options]"

    takes_optiongroups = {
        "sambaopts": options.SambaOptions,
        "credopts": options.CredentialsOptions,
        "hostopts": options.HostOptions,
    }

    takes_options = [
        Option("--json", help="Output results in JSON format.",
               dest="output_format", action="store_const", const="json"),
    ]

    def run(self, hostopts=None, sambaopts=None, credopts=None,
            output_format=None):

        ldb = self.ldb_connect(hostopts, sambaopts, credopts)

        # Value types grouped by display name.
        try:
            value_types = {value_type.display_name: value_type.as_dict()
                           for value_type in ValueType.query(ldb)}
        except ModelError as e:
            raise CommandError(e)

        # Using json output format gives more detail.
        if output_format == "json":
            self.print_json(value_types)
        else:
            for value_type in value_types.keys():
                self.outf.write(f"{value_type}\n")


class cmd_domain_claim_value_type_view(Command):
    """View a single claim value type on the domain."""

    synopsis = "%prog -H <URL> [options]"

    takes_optiongroups = {
        "sambaopts": options.SambaOptions,
        "credopts": options.CredentialsOptions,
        "hostopts": options.HostOptions,
    }

    takes_options = [
        Option("--name",
               help="Display name of claim value type to view (required).",
               dest="name", action="store", type=str, required=True),
    ]

    def run(self, hostopts=None, sambaopts=None, credopts=None, name=None):

        ldb = self.ldb_connect(hostopts, sambaopts, credopts)

        try:
            value_type = ValueType.get(ldb, display_name=name)
        except ModelError as e:
            raise CommandError(e)

        # Check if value type exists first.
        if value_type is None:
            raise CommandError(f"Value type {name} not found.")

        # Display vale type as JSON.
        self.print_json(value_type.as_dict())


class cmd_domain_claim_value_type(SuperCommand):
    """Manage claim value types on the domain."""

    subcommands = {
        "list": cmd_domain_claim_value_type_list(),
        "view": cmd_domain_claim_value_type_view(),
    }



# Unix SMB/CIFS implementation.
# Copyright (C) Andrew Bartlett <abartlet@samba.org> 2017
# Copyright (C) Catalyst IT Ltd. 2017
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
    Tests that exercise auth logging for unsuccessful netlogon attempts.

    NOTE: netlogon is only done once per session, so this file should only
          test failed logons.  Adding a successful case will potentially break
          the other tests, depending on the order of execution.
"""

import samba.tests
import os
from samba import NTSTATUSError
from samba.samdb import SamDB
import samba.tests.auth_log_base
from samba.credentials import Credentials
from samba.dcerpc import netlogon
from samba.dcerpc.dcerpc import AS_SYSTEM_MAGIC_PATH_TOKEN
from samba.auth import system_session
from samba.tests import delete_force
from samba.dsdb import UF_WORKSTATION_TRUST_ACCOUNT, UF_PASSWD_NOTREQD
from samba.dcerpc.misc import SEC_CHAN_WKSTA
from samba.dcerpc.netlogon import NETLOGON_NEG_STRONG_KEYS
from samba.common import get_string
from samba.dcerpc.windows_event_ids import (
    EVT_ID_UNSUCCESSFUL_LOGON,
    EVT_LOGON_NETWORK
)


class AuthLogTestsNetLogonBadCreds(samba.tests.auth_log_base.AuthLogTestBase):

    def setUp(self):
        super().setUp()
        self.lp = samba.tests.env_loadparm()
        self.session = system_session()
        self.ldb = SamDB(
            session_info=self.session,
            lp=self.lp)

        self.domain = os.environ["DOMAIN"]
        self.netbios_name = "NetLogonBad"
        self.machinepass = "abcdefghij"
        self.remoteAddress = AS_SYSTEM_MAGIC_PATH_TOKEN
        self.base_dn = self.ldb.domain_dn()
        self.dn = ("cn=%s,cn=users,%s" % (self.netbios_name, self.base_dn))

        utf16pw = get_string('"' + self.machinepass + '"').encode('utf-16-le')
        self.ldb.add({
            "dn": self.dn,
            "objectclass": "computer",
            "sAMAccountName": "%s$" % self.netbios_name,
            "userAccountControl":
                str(UF_WORKSTATION_TRUST_ACCOUNT | UF_PASSWD_NOTREQD),
            "unicodePwd": utf16pw})

    def tearDown(self):
        super().tearDown()
        delete_force(self.ldb, self.dn)

    def _test_netlogon(self, name, pwd, status, checkFunction, event_id):

        def isLastExpectedMessage(msg):
            return (
                msg["type"] == "Authentication" and
                msg["Authentication"]["serviceDescription"] == "NETLOGON" and
                msg["Authentication"]["authDescription"] ==
                "ServerAuthenticate" and
                msg["Authentication"]["status"] == status and
                msg["Authentication"]["eventId"] == event_id and
                msg["Authentication"]["logonType"] == EVT_LOGON_NETWORK)

        machine_creds = Credentials()
        machine_creds.guess(self.get_loadparm())
        machine_creds.set_secure_channel_type(SEC_CHAN_WKSTA)
        machine_creds.set_password(pwd)
        machine_creds.set_username(name + "$")

        try:
            netlogon.netlogon("ncalrpc:[schannel]",
                              self.get_loadparm(),
                              machine_creds)
            self.fail("NTSTATUSError not raised")
        except NTSTATUSError:
            pass

        messages = self.waitForMessages(isLastExpectedMessage)
        checkFunction(messages)

    def netlogon_check(self, messages):

        expected_messages = 4
        self.assertEqual(expected_messages,
                          len(messages),
                          "Did not receive the expected number of messages")

        # Check the first message it should be an Authorization
        msg = messages[0]
        self.assertEqual("Authorization", msg["type"])
        self.assertEqual("DCE/RPC",
                          msg["Authorization"]["serviceDescription"])
        self.assertEqual("ncalrpc", msg["Authorization"]["authType"])
        self.assertEqual("NONE", msg["Authorization"]["transportProtection"])
        self.assertTrue(self.is_guid(msg["Authorization"]["sessionId"]))

    def test_netlogon_bad_machine_name(self):
        self._test_netlogon("bad_name",
                            self.machinepass,
                            "NT_STATUS_NO_TRUST_SAM_ACCOUNT",
                            self.netlogon_check,
                            EVT_ID_UNSUCCESSFUL_LOGON)

    def test_netlogon_bad_password(self):
        self._test_netlogon(self.netbios_name,
                            "badpass",
                            "NT_STATUS_ACCESS_DENIED",
                            self.netlogon_check,
                            EVT_ID_UNSUCCESSFUL_LOGON)

    def test_netlogon_password_DES(self):
        """Logon failure that exercises the "DES" passwordType path.
        """
        def isLastExpectedMessage(msg):
            return (
                msg["type"] == "Authentication" and
                msg["Authentication"]["serviceDescription"] == "NETLOGON" and
                msg["Authentication"]["authDescription"] ==
                "ServerAuthenticate" and
                msg["Authentication"]["passwordType"] == "DES" and
                (msg["Authentication"]["eventId"] ==
                    EVT_ID_UNSUCCESSFUL_LOGON) and
                msg["Authentication"]["logonType"] == EVT_LOGON_NETWORK)

        c = netlogon.netlogon("ncalrpc:[schannel]", self.get_loadparm())
        creds = netlogon.netr_Credential()
        c.netr_ServerReqChallenge(self.server, self.netbios_name, creds)
        try:
            c.netr_ServerAuthenticate3(self.server,
                                       self.netbios_name,
                                       SEC_CHAN_WKSTA,
                                       self.netbios_name,
                                       creds,
                                       0)
        except NTSTATUSError:
            pass
        self.waitForMessages(isLastExpectedMessage)

    def test_netlogon_password_HMAC_MD5(self):
        """Logon failure that exercises the "HMAC-MD5" passwordType path.
        """
        def isLastExpectedMessage(msg):
            return (
                msg["type"] == "Authentication" and
                msg["Authentication"]["serviceDescription"] == "NETLOGON" and
                msg["Authentication"]["authDescription"] ==
                "ServerAuthenticate" and
                msg["Authentication"]["passwordType"] == "HMAC-MD5" and
                (msg["Authentication"]["eventId"] ==
                    EVT_ID_UNSUCCESSFUL_LOGON) and
                msg["Authentication"]["logonType"] == EVT_LOGON_NETWORK)

        c = netlogon.netlogon("ncalrpc:[schannel]", self.get_loadparm())
        creds = netlogon.netr_Credential()
        c.netr_ServerReqChallenge(self.server, self.netbios_name, creds)
        try:
            c.netr_ServerAuthenticate3(self.server,
                                       self.netbios_name,
                                       SEC_CHAN_WKSTA,
                                       self.netbios_name,
                                       creds,
                                       NETLOGON_NEG_STRONG_KEYS)
        except NTSTATUSError:
            pass
        self.waitForMessages(isLastExpectedMessage)



# Unix SMB/CIFS implementation.
#
# Copyright (C) David Mulder <dmulder@samba.org> 2020
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Confirm that net_s3.join_member works
"""

import samba.tests
import os
from samba.net_s3 import Net as s3_Net
from samba.credentials import DONT_USE_KERBEROS
from samba.samba3 import param as s3param
from samba import WERRORError


def rm(rmdir):
    for f in os.listdir(rmdir):
        if os.path.isdir(os.path.join(rmdir, f)):
            rm(os.path.join(rmdir, f))
            os.rmdir(os.path.join(rmdir, f))
        else:
            os.unlink(os.path.join(rmdir, f))

class NetS3JoinTests(samba.tests.TestCaseInTempDir):

    def setUp(self):
        super().setUp()
        self.realm = os.environ["REALM"]
        self.domain = os.environ["DOMAIN"]
        self.server = os.environ["SERVER"]
        self.lp = self.get_loadparm()

    def test_net_join(self):
        netbios_name = "S3NetJoinTest"
        machinepass  = "abcdefghij"
        creds = self.insta_creds(template=self.get_credentials(),
                                 kerberos_state=DONT_USE_KERBEROS)
        s3_lp = s3param.get_context()
        s3_lp.load(self.lp.configfile)

        s3_lp.set('realm', self.realm)
        s3_lp.set('workgroup', self.domain)
        s3_lp.set("private dir", self.tempdir)
        s3_lp.set("lock dir", self.tempdir)
        s3_lp.set("state directory", self.tempdir)
        s3_lp.set('server role', 'member server')
        net = s3_Net(creds, s3_lp, server=self.server)

        try:
            (domain_sid, domain_name) = net.join_member(netbios_name,
                                                        machinepass=machinepass)
        except WERRORError as e:
            self.fail('Join failed: %s' % e.args[1])
            raise

        try:
            ret = net.leave()
        except WERRORError as e:
            self.fail('Leave failed: %s' % e.args[1])
            raise
        self.assertTrue(ret, 'Leave failed!')
        rm(self.tempdir)



# domain management - domain schemaupgrade
#
# Copyright Matthias Dieter Wallnoefer 2009
# Copyright Andrew Kroeger 2009
# Copyright Jelmer Vernooij 2007-2012
# Copyright Giampaolo Lauria 2011
# Copyright Matthieu Patou <mat@matws.net> 2011
# Copyright Andrew Bartlett 2008-2015
# Copyright Stefan Metzmacher 2012
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import shutil
import subprocess
import tempfile

import ldb
import samba.getopt as options
from samba.auth import system_session
from samba.netcmd import Command, CommandError, Option
from samba.netcmd.fsmo import get_fsmo_roleowner
from samba.provision import setup_path
from samba.samdb import SamDB


class ldif_schema_update:
    """Helper class for applying LDIF schema updates"""

    def __init__(self):
        self.is_defunct = False
        self.unknown_oid = None
        self.dn = None
        self.ldif = ""

    def can_ignore_failure(self, error):
        """Checks if we can safely ignore failure to apply an LDIF update"""
        (num, errstr) = error.args

        # Microsoft has marked objects as defunct that Samba doesn't know about
        if num == ldb.ERR_NO_SUCH_OBJECT and self.is_defunct:
            print("Defunct object %s doesn't exist, skipping" % self.dn)
            return True
        elif self.unknown_oid is not None:
            print("Skipping unknown OID %s for object %s" % (self.unknown_oid, self.dn))
            return True

        return False

    def apply(self, samdb):
        """Applies a single LDIF update to the schema"""

        try:
            try:
                samdb.modify_ldif(self.ldif, controls=['relax:0'])
            except ldb.LdbError as e:
                if e.args[0] == ldb.ERR_INVALID_ATTRIBUTE_SYNTAX:

                    # REFRESH after a failed change

                    # Otherwise the OID-to-attribute mapping in
                    # _apply_updates_in_file() won't work, because it
                    # can't lookup the new OID in the schema
                    samdb.set_schema_update_now()

                    samdb.modify_ldif(self.ldif, controls=['relax:0'])
                else:
                    raise
        except ldb.LdbError as e:
            if self.can_ignore_failure(e):
                return 0
            else:
                print("Exception: %s" % e)
                print("Encountered while trying to apply the following LDIF")
                print("----------------------------------------------------")
                print("%s" % self.ldif)

                raise

        return 1


class cmd_domain_schema_upgrade(Command):
    """Domain schema upgrading"""

    synopsis = "%prog [options]"

    takes_optiongroups = {
        "sambaopts": options.SambaOptions,
        "versionopts": options.VersionOptions,
        "credopts": options.CredentialsOptions,
    }

    takes_options = [
        Option("-H", "--URL", help="LDB URL for database or target server", type=str,
               metavar="URL", dest="H"),
        Option("-q", "--quiet", help="Be quiet", action="store_true"),  # unused
        Option("-v", "--verbose", help="Be verbose", action="store_true"),
        Option("--schema", type="choice", metavar="SCHEMA",
               choices=["2012", "2012_R2", "2016", "2019"],
               help="The schema file to upgrade to. Default is (Windows) 2019.",
               default="2019"),
        Option("--ldf-file", type=str, default=None,
               help="Just apply the schema updates in the adprep/.LDF file(s) specified"),
        Option("--base-dir", type=str, default=None,
               help="Location of ldf files Default is ${SETUPDIR}/adprep.")
    ]

    def _apply_updates_in_file(self, samdb, ldif_file):
        """
        Applies a series of updates specified in an .LDIF file. The .LDIF file
        is based on the adprep Schema updates provided by Microsoft.
        """
        count = 0
        ldif_op = ldif_schema_update()

        # parse the file line by line and work out each update operation to apply
        for line in ldif_file:

            line = line.rstrip()

            # the operations in the .LDIF file are separated by blank lines. If
            # we hit a blank line, try to apply the update we've parsed so far
            if line == '':

                # keep going if we haven't parsed anything yet
                if ldif_op.ldif == '':
                    continue

                # Apply the individual change
                count += ldif_op.apply(samdb)

                # start storing the next operation from scratch again
                ldif_op = ldif_schema_update()
                continue

            # replace the placeholder domain name in the .ldif file with the real domain
            if line.upper().endswith('DC=X'):
                line = line[:-len('DC=X')] + str(samdb.get_default_basedn())
            elif line.upper().endswith('CN=X'):
                line = line[:-len('CN=X')] + str(samdb.get_default_basedn())

            values = line.split(':')

            if values[0].lower() == 'dn':
                ldif_op.dn = values[1].strip()

            # replace the Windows-specific operation with the Samba one
            if values[0].lower() == 'changetype':
                line = line.lower().replace(': ntdsschemaadd',
                                            ': add')
                line = line.lower().replace(': ntdsschemamodify',
                                            ': modify')
                line = line.lower().replace(': ntdsschemamodrdn',
                                            ': modrdn')
                line = line.lower().replace(': ntdsschemadelete',
                                            ': delete')

            if values[0].lower() in ['rdnattid', 'subclassof',
                                     'systemposssuperiors',
                                     'systemmaycontain',
                                     'systemauxiliaryclass']:
                _, value = values

                # The Microsoft updates contain some OIDs we don't recognize.
                # Query the DB to see if we can work out the OID this update is
                # referring to. If we find a match, then replace the OID with
                # the ldapDisplayname
                if '.' in value:
                    res = samdb.search(base=samdb.get_schema_basedn(),
                                       expression="(|(attributeId=%s)(governsId=%s))" %
                                       (value, value),
                                       attrs=['ldapDisplayName'])

                    if len(res) != 1:
                        ldif_op.unknown_oid = value
                    else:
                        display_name = str(res[0]['ldapDisplayName'][0])
                        line = line.replace(value, ' ' + display_name)

            # Microsoft has marked objects as defunct that Samba doesn't know about
            if values[0].lower() == 'isdefunct' and values[1].strip().lower() == 'true':
                ldif_op.is_defunct = True

            # Samba has added the showInAdvancedViewOnly attribute to all objects,
            # so rather than doing an add, we need to do a replace
            if values[0].lower() == 'add' and values[1].strip().lower() == 'showinadvancedviewonly':
                line = 'replace: showInAdvancedViewOnly'

            # Add the line to the current LDIF operation (including the newline
            # we stripped off at the start of the loop)
            ldif_op.ldif += line + '\n'

        return count

    def _apply_update(self, samdb, update_file, base_dir):
        """Wrapper function for parsing an LDIF file and applying the updates"""

        print("Applying %s updates..." % update_file)

        ldif_file = None
        try:
            ldif_file = open(os.path.join(base_dir, update_file))

            count = self._apply_updates_in_file(samdb, ldif_file)

        finally:
            if ldif_file:
                ldif_file.close()

        print("%u changes applied" % count)

        return count

    def run(self, **kwargs):
        try:
            from samba.ms_schema_markdown import read_ms_markdown
        except ImportError as e:
            self.outf.write("Exception in importing markdown: %s\n" % e)
            raise CommandError('Failed to import module markdown')
        from samba.schema import Schema

        updates_allowed_overridden = False
        sambaopts = kwargs.get("sambaopts")
        credopts = kwargs.get("credopts")
        lp = sambaopts.get_loadparm()
        creds = credopts.get_credentials(lp)
        H = kwargs.get("H")
        target_schema = kwargs.get("schema")
        ldf_files = kwargs.get("ldf_file")
        base_dir = kwargs.get("base_dir")

        temp_folder = None

        samdb = SamDB(url=H, session_info=system_session(), credentials=creds, lp=lp)

        # we're not going to get far if the config doesn't allow schema updates
        if lp.get("dsdb:schema update allowed") is None:
            lp.set("dsdb:schema update allowed", "yes")
            print("Temporarily overriding 'dsdb:schema update allowed' setting")
            updates_allowed_overridden = True

        own_dn = ldb.Dn(samdb, samdb.get_dsServiceName())
        master = get_fsmo_roleowner(samdb, str(samdb.get_schema_basedn()),
                                    'schema')
        if own_dn != master:
            raise CommandError("This server is not the schema master.")

        # if specific LDIF files were specified, just apply them
        if ldf_files:
            schema_updates = ldf_files.split(",")
        else:
            schema_updates = []

            # work out the version of the target schema we're upgrading to
            end = Schema.get_version(target_schema)

            # work out the version of the schema we're currently using
            res = samdb.search(base=samdb.get_schema_basedn(),
                               scope=ldb.SCOPE_BASE, attrs=['objectVersion'])

            if len(res) != 1:
                raise CommandError('Could not determine current schema version')
            start = int(res[0]['objectVersion'][0]) + 1

            diff_dir = setup_path("adprep/WindowsServerDocs")
            if base_dir is None:
                # Read from the Schema-Updates.md file
                temp_folder = tempfile.mkdtemp()

                update_file = setup_path("adprep/WindowsServerDocs/Schema-Updates.md")

                try:
                    read_ms_markdown(update_file, temp_folder)
                except Exception as e:
                    print("Exception in markdown parsing: %s" % e)
                    shutil.rmtree(temp_folder)
                    raise CommandError('Failed to upgrade schema')

                base_dir = temp_folder

            for version in range(start, end + 1):
                update = 'Sch%d.ldf' % version
                schema_updates.append(update)

                # Apply patches if we parsed the Schema-Updates.md file
                diff = os.path.abspath(os.path.join(diff_dir, update + '.diff'))
                if temp_folder and os.path.exists(diff):
                    try:
                        p = subprocess.Popen(['patch', update, '-i', diff],
                                             stdout=subprocess.PIPE,
                                             stderr=subprocess.PIPE, cwd=temp_folder)
                    except (OSError, IOError):
                        shutil.rmtree(temp_folder)
                        raise CommandError("Failed to upgrade schema. "
                                           "Is '/usr/bin/patch' missing?")

                    stdout, stderr = p.communicate()

                    if p.returncode:
                        print("Exception in patch: %s\n%s" % (stdout, stderr))
                        shutil.rmtree(temp_folder)
                        raise CommandError('Failed to upgrade schema')

                    print("Patched %s using %s" % (update, diff))

        if base_dir is None:
            base_dir = setup_path("adprep")

        samdb.transaction_start()
        count = 0
        error_encountered = False

        try:
            # Apply the schema updates needed to move to the new schema version
            for ldif_file in schema_updates:
                count += self._apply_update(samdb, ldif_file, base_dir)

            if count > 0:
                samdb.transaction_commit()
                print("Schema successfully updated")
            else:
                print("No changes applied to schema")
                samdb.transaction_cancel()
        except Exception as e:
            print("Exception: %s" % e)
            print("Error encountered, aborting schema upgrade")
            samdb.transaction_cancel()
            error_encountered = True

        if updates_allowed_overridden:
            lp.set("dsdb:schema update allowed", "no")

        if temp_folder:
            shutil.rmtree(temp_folder)

        if error_encountered:
            raise CommandError('Failed to upgrade schema')



# Unix SMB/CIFS implementation.
#
# Query class for the ORM to the Ldb database.
#
# Copyright (C) Catalyst.Net Ltd. 2023
#
# Written by Rob van der Linde <rob@catalyst.net.nz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import re

from .exceptions import DoesNotExist, MultipleObjectsReturned

RE_SPLIT_CAMELCASE = re.compile(r"[A-Z](?:[a-z]+|[A-Z]*(?=[A-Z]|$))")


class Query:
    """Simple Query class used by the `Model.query` method."""

    def __init__(self, model, ldb, result):
        self.model = model
        self.ldb = ldb
        self.result = result
        self.count = result.count
        self.name = " ".join(RE_SPLIT_CAMELCASE.findall(model.__name__)).lower()

    def __iter__(self):
        """Loop over Query class yields Model instances."""
        for message in self.result:
            yield self.model.from_message(self.ldb, message)

    def first(self):
        """Returns the first item in the Query or None for no results."""
        if self.result.count:
            return self.model.from_message(self.ldb, self.result[0])

    def last(self):
        """Returns the last item in the Query or None for no results."""
        if self.result.count:
            return self.model.from_message(self.ldb, self.result[-1])

    def get(self):
        """Returns one item or None if no results were found.

        :returns: Model instance or None if not found.
        :raises MultipleObjectsReturned: if more than one results were returned
        """
        if self.count > 1:
            raise MultipleObjectsReturned(
                f"More than one {self.name} objects returned (got {self.count}).")
        elif self.count:
            return self.model.from_message(self.ldb, self.result[0])

    def one(self):
        """Must return EXACTLY one item or raise an exception.

        :returns: Model instance
        :raises DoesNotExist: if no results were returned
        :raises MultipleObjectsReturned: if more than one results were returned
        """
        if self.count < 1:
            raise DoesNotExist(
                f"{self.name.capitalize()} matching query not found")
        elif self.count > 1:
            raise MultipleObjectsReturned(
                f"More than one {self.name} objects returned (got {self.count}).")
        else:
            return self.model.from_message(self.ldb, self.result[0])



# Test 'samba-tool domain passwordsettings' sub-commands
#
# Copyright (C) Andrew Bartlett <abartlet@samba.org> 2018
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import ldb
from samba.tests.samba_tool.base import SambaToolCmdTest
from samba.tests.pso import PasswordSettings, TestUser


class PwdSettingsCmdTestCase(SambaToolCmdTest):
    """Tests for 'samba-tool domain passwordsettings' subcommands"""

    def setUp(self):
        super().setUp()
        self.server = "ldap://%s" % os.environ["DC_SERVER"]
        self.user_auth = "-U%s%%%s" % (os.environ["DC_USERNAME"],
                                       os.environ["DC_PASSWORD"])
        self.ldb = self.getSamDB("-H", self.server, self.user_auth)
        system_dn = "CN=System,%s" % self.ldb.domain_dn()
        self.pso_container = "CN=Password Settings Container,%s" % system_dn
        self.obj_cleanup = []

    def tearDown(self):
        super().tearDown()
        # clean-up any objects the test has created
        for dn in self.obj_cleanup:
            self.ldb.delete(dn)

    def check_pso(self, pso_name, pso):
        """Checks the PSO info in the DB matches what's expected"""

        # lookup the PSO in the DB
        dn = "CN=%s,%s" % (pso_name, self.pso_container)
        pso_attrs = ['name', 'msDS-PasswordSettingsPrecedence',
                     'msDS-PasswordReversibleEncryptionEnabled',
                     'msDS-PasswordHistoryLength',
                     'msDS-MinimumPasswordLength',
                     'msDS-PasswordComplexityEnabled',
                     'msDS-MinimumPasswordAge',
                     'msDS-MaximumPasswordAge',
                     'msDS-LockoutObservationWindow',
                     'msDS-LockoutThreshold', 'msDS-LockoutDuration']
        res = self.ldb.search(dn, scope=ldb.SCOPE_BASE, attrs=pso_attrs)
        self.assertEqual(len(res), 1, "PSO lookup failed")

        # convert types in the PSO-settings to what the search returns, i.e.
        # boolean --> string, seconds --> timestamps in -100 nanosecond units
        complexity_str = "TRUE" if pso.complexity else "FALSE"
        plaintext_str = "TRUE" if pso.store_plaintext else "FALSE"
        lockout_duration = -int(pso.lockout_duration * (1e7))
        lockout_window = -int(pso.lockout_window * (1e7))
        min_age = -int(pso.password_age_min * (1e7))
        max_age = -int(pso.password_age_max * (1e7))

        # check the PSO's settings match the search results
        self.assertEqual(str(res[0]['msDS-PasswordComplexityEnabled'][0]),
                          complexity_str)
        plaintext_res = res[0]['msDS-PasswordReversibleEncryptionEnabled'][0]
        self.assertEqual(str(plaintext_res), plaintext_str)
        self.assertEqual(int(res[0]['msDS-PasswordHistoryLength'][0]),
                          pso.history_len)
        self.assertEqual(int(res[0]['msDS-MinimumPasswordLength'][0]),
                          pso.password_len)
        self.assertEqual(int(res[0]['msDS-MinimumPasswordAge'][0]), min_age)
        self.assertEqual(int(res[0]['msDS-MaximumPasswordAge'][0]), max_age)
        self.assertEqual(int(res[0]['msDS-LockoutObservationWindow'][0]),
                          lockout_window)
        self.assertEqual(int(res[0]['msDS-LockoutDuration'][0]),
                          lockout_duration)
        self.assertEqual(int(res[0]['msDS-LockoutThreshold'][0]),
                          pso.lockout_attempts)
        self.assertEqual(int(res[0]['msDS-PasswordSettingsPrecedence'][0]),
                          pso.precedence)

        # check we can also display the PSO via the show command
        (result, out, err) = self.runsublevelcmd("domain", ("passwordsettings",
                                                 "pso", "show"), pso_name,
                                                 "-H", self.server,
                                                 self.user_auth)
        self.assertTrue(len(out.split(":")) >= 10,
                        "Expect 10 fields displayed")

        # for a few settings, sanity-check the display is what we expect
        self.assertIn("Minimum password length: %u" % pso.password_len, out)
        self.assertIn("Password history length: %u" % pso.history_len, out)
        lockout_str = "lockout threshold (attempts): %u" % pso.lockout_attempts
        self.assertIn(lockout_str, out)

    def test_pso_create(self):
        """Tests basic PSO creation using the samba-tool"""

        # we expect the PSO to take the current domain settings by default
        # (we'll set precedence/complexity, the rest should be the defaults)
        expected_pso = PasswordSettings(None, self.ldb)
        expected_pso.complexity = False
        expected_pso.precedence = 100

        # check basic PSO creation works
        pso_name = "test-create-PSO"
        (result, out, err) = self.runsublevelcmd("domain", ("passwordsettings",
                                                 "pso", "create"), pso_name,
                                                 "100", "--complexity=off",
                                                 "-H", self.server,
                                                 self.user_auth)
        # make sure we clean-up after the test completes
        self.obj_cleanup.append("CN=%s,%s" % (pso_name, self.pso_container))

        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        self.assertIn("successfully created", out)
        self.check_pso(pso_name, expected_pso)

        # check creating a PSO with the same name fails
        (result, out, err) = self.runsublevelcmd("domain", ("passwordsettings",
                                                 "pso", "create"), pso_name,
                                                 "100", "--complexity=off",
                                                 "-H", self.server,
                                                 self.user_auth)
        self.assertCmdFail(result, "Ensure that create for existing PSO fails")
        self.assertIn("already exists", err)

        # check we need to specify at least one password policy argument
        pso_name = "test-create-PSO2"
        (result, out, err) = self.runsublevelcmd("domain", ("passwordsettings",
                                                 "pso", "create"), pso_name,
                                                 "100", "-H", self.server,
                                                 self.user_auth)
        self.assertCmdFail(result, "Ensure that create for existing PSO fails")
        self.assertIn("specify at least one password policy setting", err)

        # create a PSO with different settings and check they match
        expected_pso.complexity = True
        expected_pso.store_plaintext = True
        expected_pso.precedence = 50
        expected_pso.password_len = 12
        day_in_secs = 60 * 60 * 24
        expected_pso.password_age_min = 11 * day_in_secs
        expected_pso.password_age_max = 50 * day_in_secs

        (result, out, err) = self.runsublevelcmd("domain", ("passwordsettings",
                                                 "pso", "create"), pso_name,
                                                 "50", "--complexity=on",
                                                 "--store-plaintext=on",
                                                 "--min-pwd-length=12",
                                                 "--min-pwd-age=11",
                                                 "--max-pwd-age=50",
                                                 "-H", self.server,
                                                 self.user_auth)
        self.obj_cleanup.append("CN=%s,%s" % (pso_name, self.pso_container))
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        self.assertIn("successfully created", out)
        self.check_pso(pso_name, expected_pso)

        # check the PSOs we created are present in the 'list' command
        (result, out, err) = self.runsublevelcmd("domain", ("passwordsettings",
                                                 "pso", "list"),
                                                 "-H", self.server,
                                                 self.user_auth)
        self.assertCmdSuccess(result, out, err)
        self.assertIn("test-create-PSO", out)
        self.assertIn("test-create-PSO2", out)

    def _create_pso(self, pso_name):
        """Creates a PSO for use in other tests"""
        # the new PSO will take the current domain settings by default
        pso_settings = PasswordSettings(None, self.ldb)
        pso_settings.name = pso_name
        pso_settings.password_len = 10
        pso_settings.precedence = 200

        (result, out, err) = self.runsublevelcmd("domain", ("passwordsettings",
                                                 "pso", "create"), pso_name,
                                                 "200", "--min-pwd-length=10",
                                                 "-H", self.server,
                                                 self.user_auth)
        # make sure we clean-up after the test completes
        pso_settings.dn = "CN=%s,%s" % (pso_name, self.pso_container)
        self.obj_cleanup.append(pso_settings.dn)

        # sanity-check the cmd was successful
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        self.assertIn("successfully created", out)
        self.check_pso(pso_name, pso_settings)

        return pso_settings

    def test_pso_set(self):
        """Tests we can modify a PSO using the samba-tool"""

        pso_name = "test-set-PSO"
        pso_settings = self._create_pso(pso_name)

        # check we can update a PSO's settings
        pso_settings.precedence = 99
        pso_settings.lockout_attempts = 10
        pso_settings.lockout_duration = 60 * 17
        (res, out, err) = self.runsublevelcmd("domain", ("passwordsettings",
                                              "pso", "set"), pso_name,
                                              "--precedence=99",
                                              "--account-lockout-threshold=10",
                                              "--account-lockout-duration=17",
                                              "-H", self.server,
                                              self.user_auth)
        self.assertCmdSuccess(res, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        self.assertIn("Successfully updated", out)

        # check the PSO's settings now reflect the new values
        self.check_pso(pso_name, pso_settings)

    def test_pso_delete(self):
        """Tests we can delete a PSO using the samba-tool"""

        pso_name = "test-delete-PSO"
        self._create_pso(pso_name)

        # check we can successfully delete the PSO
        (result, out, err) = self.runsublevelcmd("domain", ("passwordsettings",
                                                 "pso", "delete"), pso_name,
                                                 "-H", self.server,
                                                 self.user_auth)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        self.assertIn("Deleted PSO", out)
        dn = "CN=%s,%s" % (pso_name, self.pso_container)
        self.obj_cleanup.remove(dn)

        # check the object no longer exists in the DB
        try:
            self.ldb.search(dn, scope=ldb.SCOPE_BASE, attrs=['name'])
            self.fail("PSO shouldn't exist")
        except ldb.LdbError as e:
            (enum, estr) = e.args
            self.assertEqual(enum, ldb.ERR_NO_SUCH_OBJECT)

        # run the same cmd again - it should fail because PSO no longer exists
        (result, out, err) = self.runsublevelcmd("domain", ("passwordsettings",
                                                 "pso", "delete"), pso_name,
                                                 "-H", self.server,
                                                 self.user_auth)
        self.assertCmdFail(result, "Deleting a non-existent PSO should fail")
        self.assertIn("Unable to find PSO", err)

    def check_pso_applied(self, user, pso):
        """Checks that the correct PSO is applied to a given user"""

        # first check the samba-tool output tells us the correct PSO is applied
        (result, out, err) = self.runsublevelcmd("domain", ("passwordsettings",
                                                 "pso", "show-user"),
                                                 user.name, "-H", self.server,
                                                 self.user_auth)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        if pso is None:
            self.assertIn("No PSO applies to user", out)
        else:
            self.assertIn(pso.name, out)

        # then check the DB tells us the same thing
        if pso is None:
            self.assertEqual(user.get_resultant_PSO(), None)
        else:
            self.assertEqual(user.get_resultant_PSO(), pso.dn)

    def test_pso_apply_to_user(self):
        """Checks we can apply/unapply a PSO to a user"""

        pso_name = "test-apply-PSO"
        test_pso = self._create_pso(pso_name)

        # check that a new user has no PSO applied by default
        user = TestUser("test-PSO-user", self.ldb)
        self.obj_cleanup.append(user.dn)
        self.check_pso_applied(user, pso=None)

        # add the user to a new group
        group_name = "test-PSO-group"
        dn = "CN=%s,%s" % (group_name, self.ldb.domain_dn())
        self.ldb.add({"dn": dn, "objectclass": "group",
                      "sAMAccountName": group_name})
        self.obj_cleanup.append(dn)
        m = ldb.Message()
        m.dn = ldb.Dn(self.ldb, dn)
        m["member"] = ldb.MessageElement(user.dn, ldb.FLAG_MOD_ADD, "member")
        self.ldb.modify(m)

        # check samba-tool can successfully link a PSO to a group
        (result, out, err) = self.runsublevelcmd("domain", ("passwordsettings",
                                                 "pso", "apply"), pso_name,
                                                 group_name, "-H", self.server,
                                                 self.user_auth)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        self.check_pso_applied(user, pso=test_pso)

        # we should fail if we try to apply the same PSO/group twice though
        (result, out, err) = self.runsublevelcmd("domain", ("passwordsettings",
                                                 "pso", "apply"), pso_name,
                                                 group_name, "-H", self.server,
                                                 self.user_auth)
        self.assertCmdFail(result, "Shouldn't be able to apply PSO twice")
        self.assertIn("already applies", err)

        # check samba-tool can successfully link a PSO to a user
        (result, out, err) = self.runsublevelcmd("domain", ("passwordsettings",
                                                 "pso", "apply"), pso_name,
                                                 user.name, "-H", self.server,
                                                 self.user_auth)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        self.check_pso_applied(user, pso=test_pso)

        # check samba-tool can successfully unlink a group from a PSO
        (result, out, err) = self.runsublevelcmd("domain", ("passwordsettings",
                                                 "pso", "unapply"), pso_name,
                                                 group_name, "-H", self.server,
                                                 self.user_auth)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        # PSO still applies directly to the user, even though group was removed
        self.check_pso_applied(user, pso=test_pso)

        # check samba-tool can successfully unlink a user from a PSO
        (result, out, err) = self.runsublevelcmd("domain", ("passwordsettings",
                                                 "pso", "unapply"), pso_name,
                                                 user.name, "-H", self.server,
                                                 self.user_auth)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        self.check_pso_applied(user, pso=None)

    def test_pso_unpriv(self):
        """Checks unprivileged users can't modify PSOs via samba-tool"""

        # create a dummy PSO and a non-admin user
        pso_name = "test-unpriv-PSO"
        self._create_pso(pso_name)
        user = TestUser("test-unpriv-user", self.ldb)
        self.obj_cleanup.append(user.dn)
        unpriv_auth = "-U%s%%%s" % (user.name, user.get_password())

        # check we need admin privileges to be able to do anything to PSOs
        (result, out, err) = self.runsublevelcmd("domain", ("passwordsettings",
                                                 "pso", "set"), pso_name,
                                                 "--complexity=off", "-H",
                                                 self.server, unpriv_auth)
        self.assertCmdFail(result, "Need admin privileges to modify PSO")
        self.assertIn("You may not have permission", err)

        (result, out, err) = self.runsublevelcmd("domain", ("passwordsettings",
                                                 "pso", "create"), "bad-perm",
                                                 "250", "--complexity=off",
                                                 "-H", self.server,
                                                 unpriv_auth)
        self.assertCmdFail(result, "Need admin privileges to modify PSO")
        self.assertIn("Administrator permissions are needed", err)

        (result, out, err) = self.runsublevelcmd("domain", ("passwordsettings",
                                                 "pso", "delete"), pso_name,
                                                 "-H", self.server,
                                                 unpriv_auth)
        self.assertCmdFail(result, "Need admin privileges to delete PSO")
        self.assertIn("You may not have permission", err)

        (result, out, err) = self.runsublevelcmd("domain", ("passwordsettings",
                                                 "pso", "show"), pso_name,
                                                 "-H", self.server,
                                                 unpriv_auth)
        self.assertCmdFail(result, "Need admin privileges to view PSO")
        self.assertIn("You may not have permission", err)

        (result, out, err) = self.runsublevelcmd("domain", ("passwordsettings",
                                                 "pso", "apply"), pso_name,
                                                 user.name, "-H", self.server,
                                                 unpriv_auth)
        self.assertCmdFail(result, "Need admin privileges to modify PSO")
        self.assertIn("You may not have permission", err)

        (result, out, err) = self.runsublevelcmd("domain", ("passwordsettings",
                                                 "pso", "unapply"), pso_name,
                                                 user.name, "-H", self.server,
                                                 unpriv_auth)
        self.assertCmdFail(result, "Need admin privileges to modify PSO")
        self.assertIn("You may not have permission", err)

        # The 'list' command actually succeeds because it's not easy to tell
        # whether we got no results due to lack of permissions, or because
        # there were no PSOs to display
        (result, out, err) = self.runsublevelcmd("domain", ("passwordsettings",
                                                 "pso", "list"), "-H",
                                                 self.server, unpriv_auth)
        self.assertCmdSuccess(result, out, err)
        self.assertIn("No PSOs", out)
        self.assertIn("permission", out)

    def test_domain_passwordsettings(self):
        """Checks the 'set/show' commands for the domain settings (non-PSO)"""

        # check the 'show' cmd for the domain settings
        (result, out, err) = self.runsublevelcmd("domain", ("passwordsettings",
                                                 "show"), "-H", self.server,
                                                 self.user_auth)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")

        # check an arbitrary setting is displayed correctly
        min_pwd_len = self.ldb.get_minPwdLength()
        self.assertIn("Minimum password length: %s" % min_pwd_len, out)

        # check we can change the domain setting
        self.addCleanup(self.ldb.set_minPwdLength, min_pwd_len)
        new_len = int(min_pwd_len) + 3
        min_pwd_args = "--min-pwd-length=%u" % new_len
        (result, out, err) = self.runsublevelcmd("domain", ("passwordsettings",
                                                 "set"), min_pwd_args,
                                                 "-H", self.server,
                                                 self.user_auth)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        self.assertIn("successful", out)
        self.assertEqual(new_len, self.ldb.get_minPwdLength())

        # check the updated value is now displayed
        (result, out, err) = self.runsublevelcmd("domain", ("passwordsettings",
                                                 "show"), "-H", self.server,
                                                 self.user_auth)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        self.assertIn("Minimum password length: %u" % new_len, out)

    def test_domain_passwordsettings_pwdage(self):
        """Checks the 'set' command for the domain password age (non-PSO)"""

        # check we can set the domain max password age
        max_pwd_age = self.ldb.get_maxPwdAge()
        self.addCleanup(self.ldb.set_maxPwdAge, max_pwd_age)
        max_pwd_args = "--max-pwd-age=270"
        (result, out, err) = self.runsublevelcmd("domain", ("passwordsettings",
                                                 "set"), max_pwd_args,
                                                 "-H", self.server,
                                                 self.user_auth)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        self.assertIn("successful", out)
        self.assertNotEqual(max_pwd_age, self.ldb.get_maxPwdAge())

        # check we can't set the domain min password age to more than the max
        min_pwd_age = self.ldb.get_minPwdAge()
        self.addCleanup(self.ldb.set_minPwdAge, min_pwd_age)
        min_pwd_args = "--min-pwd-age=271"
        (result, out, err) = self.runsublevelcmd("domain", ("passwordsettings",
                                                 "set"), min_pwd_args,
                                                 "-H", self.server,
                                                 self.user_auth)
        self.assertCmdFail(result, "minPwdAge > maxPwdAge should be rejected")
        self.assertIn("Maximum password age", err)

        # check we can set the domain min password age to less than the max
        min_pwd_args = "--min-pwd-age=269"
        (result, out, err) = self.runsublevelcmd("domain", ("passwordsettings",
                                                 "set"), min_pwd_args,
                                                 "-H", self.server,
                                                 self.user_auth)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        self.assertIn("successful", out)
        self.assertNotEqual(min_pwd_age, self.ldb.get_minPwdAge())



# Reads important GPO parameters and updates Samba
# Copyright (C) Luke Morrison <luc785@.hotmail.com> 2013
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys
import os, shutil
import errno
import tdb
import pwd
sys.path.insert(0, "bin/python")
from samba import WERRORError
from configparser import ConfigParser
from io import StringIO
import traceback
from samba.common import get_bytes
from abc import ABCMeta, abstractmethod
import xml.etree.ElementTree as etree
import re
from samba.net import Net
from samba.dcerpc import nbt
from samba.samba3 import libsmb_samba_internal as libsmb
import samba.gpo as gpo
from uuid import UUID
from tempfile import NamedTemporaryFile
from samba.dcerpc import preg
from samba.ndr import ndr_unpack
from samba.credentials import SMB_SIGNING_REQUIRED
from samba.gp.util.logging import log
from hashlib import blake2b
import numbers
from samba.common import get_string
from samba.samdb import SamDB
from samba.auth import system_session
import ldb
from samba.dsdb import UF_WORKSTATION_TRUST_ACCOUNT, UF_SERVER_TRUST_ACCOUNT, GPLINK_OPT_ENFORCE, GPLINK_OPT_DISABLE, GPO_BLOCK_INHERITANCE
from samba.auth import AUTH_SESSION_INFO_DEFAULT_GROUPS, AUTH_SESSION_INFO_AUTHENTICATED, AUTH_SESSION_INFO_SIMPLE_PRIVILEGES
from samba.dcerpc import security
import samba.security
from samba.dcerpc import nbt
from datetime import datetime


try:
    from enum import Enum
    GPOSTATE = Enum('GPOSTATE', 'APPLY ENFORCE UNAPPLY')
except ImportError:
    class GPOSTATE:
        APPLY = 1
        ENFORCE = 2
        UNAPPLY = 3


class gp_log:
    """ Log settings overwritten by gpo apply
    The gp_log is an xml file that stores a history of gpo changes (and the
    original setting value).

    The log is organized like so:

<gp>
    <user name="KDC-1$">
        <applylog>
            <guid count="0" value="{31B2F340-016D-11D2-945F-00C04FB984F9}" />
        </applylog>
        <guid value="{31B2F340-016D-11D2-945F-00C04FB984F9}">
            <gp_ext name="System Access">
                <attribute name="minPwdAge">-864000000000</attribute>
                <attribute name="maxPwdAge">-36288000000000</attribute>
                <attribute name="minPwdLength">7</attribute>
                <attribute name="pwdProperties">1</attribute>
            </gp_ext>
            <gp_ext name="Kerberos Policy">
                <attribute name="ticket_lifetime">1d</attribute>
                <attribute name="renew_lifetime" />
                <attribute name="clockskew">300</attribute>
            </gp_ext>
        </guid>
    </user>
</gp>

    Each guid value contains a list of extensions, which contain a list of
    attributes. The guid value represents a GPO. The attributes are the values
    of those settings prior to the application of the GPO.
    The list of guids is enclosed within a user name, which represents the user
    the settings were applied to. This user may be the samaccountname of the
    local computer, which implies that these are machine policies.
    The applylog keeps track of the order in which the GPOs were applied, so
    that they can be rolled back in reverse, returning the machine to the state
    prior to policy application.
    """
    def __init__(self, user, gpostore, db_log=None):
        """ Initialize the gp_log
        param user          - the username (or machine name) that policies are
                              being applied to
        param gpostore      - the GPOStorage obj which references the tdb which
                              contains gp_logs
        param db_log        - (optional) a string to initialize the gp_log
        """
        self._state = GPOSTATE.APPLY
        self.gpostore = gpostore
        self.username = user
        if db_log:
            self.gpdb = etree.fromstring(db_log)
        else:
            self.gpdb = etree.Element('gp')
        self.user = user
        user_obj = self.gpdb.find('user[@name="%s"]' % user)
        if user_obj is None:
            user_obj = etree.SubElement(self.gpdb, 'user')
            user_obj.attrib['name'] = user

    def state(self, value):
        """ Policy application state
        param value         - APPLY, ENFORCE, or UNAPPLY

        The behavior of the gp_log depends on whether we are applying policy,
        enforcing policy, or unapplying policy. During an apply, old settings
        are recorded in the log. During an enforce, settings are being applied
        but the gp_log does not change. During an unapply, additions to the log
        should be ignored (since function calls to apply settings are actually
        reverting policy), but removals from the log are allowed.
        """
        # If we're enforcing, but we've unapplied, apply instead
        if value == GPOSTATE.ENFORCE:
            user_obj = self.gpdb.find('user[@name="%s"]' % self.user)
            apply_log = user_obj.find('applylog')
            if apply_log is None or len(apply_log) == 0:
                self._state = GPOSTATE.APPLY
            else:
                self._state = value
        else:
            self._state = value

    def get_state(self):
        """Check the GPOSTATE
        """
        return self._state

    def set_guid(self, guid):
        """ Log to a different GPO guid
        param guid          - guid value of the GPO from which we're applying
                              policy
        """
        self.guid = guid
        user_obj = self.gpdb.find('user[@name="%s"]' % self.user)
        obj = user_obj.find('guid[@value="%s"]' % guid)
        if obj is None:
            obj = etree.SubElement(user_obj, 'guid')
            obj.attrib['value'] = guid
        if self._state == GPOSTATE.APPLY:
            apply_log = user_obj.find('applylog')
            if apply_log is None:
                apply_log = etree.SubElement(user_obj, 'applylog')
            prev = apply_log.find('guid[@value="%s"]' % guid)
            if prev is None:
                item = etree.SubElement(apply_log, 'guid')
                item.attrib['count'] = '%d' % (len(apply_log) - 1)
                item.attrib['value'] = guid

    def store(self, gp_ext_name, attribute, old_val):
        """ Store an attribute in the gp_log
        param gp_ext_name   - Name of the extension applying policy
        param attribute     - The attribute being modified
        param old_val       - The value of the attribute prior to policy
                              application
        """
        if self._state == GPOSTATE.UNAPPLY or self._state == GPOSTATE.ENFORCE:
            return None
        user_obj = self.gpdb.find('user[@name="%s"]' % self.user)
        guid_obj = user_obj.find('guid[@value="%s"]' % self.guid)
        assert guid_obj is not None, "gpo guid was not set"
        ext = guid_obj.find('gp_ext[@name="%s"]' % gp_ext_name)
        if ext is None:
            ext = etree.SubElement(guid_obj, 'gp_ext')
            ext.attrib['name'] = gp_ext_name
        attr = ext.find('attribute[@name="%s"]' % attribute)
        if attr is None:
            attr = etree.SubElement(ext, 'attribute')
            attr.attrib['name'] = attribute
            attr.text = old_val

    def retrieve(self, gp_ext_name, attribute):
        """ Retrieve a stored attribute from the gp_log
        param gp_ext_name   - Name of the extension which applied policy
        param attribute     - The attribute being retrieved
        return              - The value of the attribute prior to policy
                              application
        """
        user_obj = self.gpdb.find('user[@name="%s"]' % self.user)
        guid_obj = user_obj.find('guid[@value="%s"]' % self.guid)
        assert guid_obj is not None, "gpo guid was not set"
        ext = guid_obj.find('gp_ext[@name="%s"]' % gp_ext_name)
        if ext is not None:
            attr = ext.find('attribute[@name="%s"]' % attribute)
            if attr is not None:
                return attr.text
        return None

    def retrieve_all(self, gp_ext_name):
        """ Retrieve all stored attributes for this user, GPO guid, and CSE
        param gp_ext_name   - Name of the extension which applied policy
        return              - The values of the attributes prior to policy
                              application
        """
        user_obj = self.gpdb.find('user[@name="%s"]' % self.user)
        guid_obj = user_obj.find('guid[@value="%s"]' % self.guid)
        assert guid_obj is not None, "gpo guid was not set"
        ext = guid_obj.find('gp_ext[@name="%s"]' % gp_ext_name)
        if ext is not None:
            attrs = ext.findall('attribute')
            return {attr.attrib['name']: attr.text for attr in attrs}
        return {}

    def get_applied_guids(self):
        """ Return a list of applied ext guids
        return              - List of guids for gpos that have applied settings
                              to the system.
        """
        guids = []
        user_obj = self.gpdb.find('user[@name="%s"]' % self.user)
        if user_obj is not None:
            apply_log = user_obj.find('applylog')
            if apply_log is not None:
                guid_objs = apply_log.findall('guid[@count]')
                guids_by_count = [(g.get('count'), g.get('value'))
                                  for g in guid_objs]
                guids_by_count.sort(reverse=True)
                guids.extend(guid for count, guid in guids_by_count)
        return guids

    def get_applied_settings(self, guids):
        """ Return a list of applied ext guids
        return              - List of tuples containing the guid of a gpo, then
                              a dictionary of policies and their values prior
                              policy application. These are sorted so that the
                              most recently applied settings are removed first.
        """
        ret = []
        user_obj = self.gpdb.find('user[@name="%s"]' % self.user)
        for guid in guids:
            guid_settings = user_obj.find('guid[@value="%s"]' % guid)
            exts = guid_settings.findall('gp_ext')
            settings = {}
            for ext in exts:
                attr_dict = {}
                attrs = ext.findall('attribute')
                for attr in attrs:
                    attr_dict[attr.attrib['name']] = attr.text
                settings[ext.attrib['name']] = attr_dict
            ret.append((guid, settings))
        return ret

    def delete(self, gp_ext_name, attribute):
        """ Remove an attribute from the gp_log
        param gp_ext_name   - name of extension from which to remove the
                              attribute
        param attribute     - attribute to remove
        """
        user_obj = self.gpdb.find('user[@name="%s"]' % self.user)
        guid_obj = user_obj.find('guid[@value="%s"]' % self.guid)
        assert guid_obj is not None, "gpo guid was not set"
        ext = guid_obj.find('gp_ext[@name="%s"]' % gp_ext_name)
        if ext is not None:
            attr = ext.find('attribute[@name="%s"]' % attribute)
            if attr is not None:
                ext.remove(attr)
                if len(ext) == 0:
                    guid_obj.remove(ext)

    def commit(self):
        """ Write gp_log changes to disk """
        self.gpostore.store(self.username, etree.tostring(self.gpdb, 'utf-8'))


class GPOStorage:
    def __init__(self, log_file):
        if os.path.isfile(log_file):
            self.log = tdb.open(log_file)
        else:
            self.log = tdb.Tdb(log_file, 0, tdb.DEFAULT, os.O_CREAT | os.O_RDWR)

    def start(self):
        self.log.transaction_start()

    def get_int(self, key):
        try:
            return int(self.log.get(get_bytes(key)))
        except TypeError:
            return None

    def get(self, key):
        return self.log.get(get_bytes(key))

    def get_gplog(self, user):
        return gp_log(user, self, self.log.get(get_bytes(user)))

    def store(self, key, val):
        self.log.store(get_bytes(key), get_bytes(val))

    def cancel(self):
        self.log.transaction_cancel()

    def delete(self, key):
        self.log.delete(get_bytes(key))

    def commit(self):
        self.log.transaction_commit()

    def __del__(self):
        self.log.close()


class gp_ext(object):
    __metaclass__ = ABCMeta

    def __init__(self, lp, creds, username, store):
        self.lp = lp
        self.creds = creds
        self.username = username
        self.gp_db = store.get_gplog(username)

    @abstractmethod
    def process_group_policy(self, deleted_gpo_list, changed_gpo_list):
        pass

    @abstractmethod
    def read(self, policy):
        pass

    def parse(self, afile):
        local_path = self.lp.cache_path('gpo_cache')
        data_file = os.path.join(local_path, check_safe_path(afile).upper())
        if os.path.exists(data_file):
            return self.read(data_file)
        return None

    @abstractmethod
    def __str__(self):
        pass

    @abstractmethod
    def rsop(self, gpo):
        return {}


class gp_inf_ext(gp_ext):
    def read(self, data_file):
        with open(data_file, 'rb') as f:
            policy = f.read()
        inf_conf = ConfigParser(interpolation=None)
        inf_conf.optionxform = str
        try:
            inf_conf.read_file(StringIO(policy.decode()))
        except UnicodeDecodeError:
            inf_conf.read_file(StringIO(policy.decode('utf-16')))
        return inf_conf


class gp_pol_ext(gp_ext):
    def read(self, data_file):
        with open(data_file, 'rb') as f:
            raw = f.read()
        return ndr_unpack(preg.file, raw)


class gp_xml_ext(gp_ext):
    def read(self, data_file):
        with open(data_file, 'rb') as f:
            raw = f.read()
        try:
            return etree.fromstring(raw.decode())
        except UnicodeDecodeError:
            return etree.fromstring(raw.decode('utf-16'))


class gp_applier(object):
    """Group Policy Applier/Unapplier/Modifier
    The applier defines functions for monitoring policy application,
    removal, and modification. It must be a multi-derived class paired
    with a subclass of gp_ext.
    """
    __metaclass__ = ABCMeta

    def cache_add_attribute(self, guid, attribute, value):
        """Add an attribute and value to the Group Policy cache
        guid        - The GPO guid which applies this policy
        attribute   - The attribute name of the policy being applied
        value       - The value of the policy being applied

        Normally called by the subclass apply() function after applying policy.
        """
        self.gp_db.set_guid(guid)
        self.gp_db.store(str(self), attribute, value)
        self.gp_db.commit()

    def cache_remove_attribute(self, guid, attribute):
        """Remove an attribute from the Group Policy cache
        guid        - The GPO guid which applies this policy
        attribute   - The attribute name of the policy being unapplied

        Normally called by the subclass unapply() function when removing old
        policy.
        """
        self.gp_db.set_guid(guid)
        self.gp_db.delete(str(self), attribute)
        self.gp_db.commit()

    def cache_get_attribute_value(self, guid, attribute):
        """Retrieve the value stored in the cache for the given attribute
        guid        - The GPO guid which applies this policy
        attribute   - The attribute name of the policy
        """
        self.gp_db.set_guid(guid)
        return self.gp_db.retrieve(str(self), attribute)

    def cache_get_all_attribute_values(self, guid):
        """Retrieve all attribute/values currently stored for this gpo+policy
        guid        - The GPO guid which applies this policy
        """
        self.gp_db.set_guid(guid)
        return self.gp_db.retrieve_all(str(self))

    def cache_get_apply_state(self):
        """Return the current apply state
        return      - APPLY|ENFORCE|UNAPPLY
        """
        return self.gp_db.get_state()

    def generate_attribute(self, name, *args):
        """Generate an attribute name from arbitrary data
        name            - A name to ensure uniqueness
        args            - Any arbitrary set of args, str or bytes
        return          - A blake2b digest of the data, the attribute

        The importance here is the digest of the data makes the attribute
        reproducible and uniquely identifies it. Hashing the name with
        the data ensures we don't falsely identify a match which is the same
        text in a different file. Using this attribute generator is optional.
        """
        data = b''.join([get_bytes(arg) for arg in [*args]])
        return blake2b(get_bytes(name)+data).hexdigest()

    def generate_value_hash(self, *args):
        """Generate a unique value which identifies value changes
        args            - Any arbitrary set of args, str or bytes
        return          - A blake2b digest of the data, the value represented
        """
        data = b''.join([get_bytes(arg) for arg in [*args]])
        return blake2b(data).hexdigest()

    @abstractmethod
    def unapply(self, guid, attribute, value):
        """Group Policy Unapply
        guid            - The GPO guid which applies this policy
        attribute       - The attribute name of the policy being unapplied
        value           - The value of the policy being unapplied
        """
        pass

    @abstractmethod
    def apply(self, guid, attribute, applier_func, *args):
        """Group Policy Apply
        guid            - The GPO guid which applies this policy
        attribute       - The attribute name of the policy being applied
        applier_func    - An applier function which takes variable args
        args            - The variable arguments to pass to applier_func

        The applier_func function MUST return the value of the policy being
        applied. It's important that implementations of `apply` check for and
        first unapply any changed policy. See for example calls to
        `cache_get_all_attribute_values()` which searches for all policies
        applied by this GPO for this Client Side Extension (CSE).
        """
        pass

    def clean(self, guid, keep=None, remove=None, **kwargs):
        """Cleanup old removed attributes
        keep    - A list of attributes to keep
        remove  - A single attribute to remove, or a list of attributes to
                  remove
        kwargs  - Additional keyword args required by the subclass unapply
                  function

        This is only necessary for CSEs which provide multiple attributes.
        """
        # Clean syntax is, either provide a single remove attribute,
        # or a list of either removal attributes or keep attributes.
        if keep is None:
            keep = []
        if remove is None:
            remove = []

        if type(remove) != list:
            value = self.cache_get_attribute_value(guid, remove)
            if value is not None:
                self.unapply(guid, remove, value, **kwargs)
        else:
            old_vals = self.cache_get_all_attribute_values(guid)
            for attribute, value in old_vals.items():
                if (len(remove) > 0 and attribute in remove) or \
                   (len(keep) > 0 and attribute not in keep):
                    self.unapply(guid, attribute, value, **kwargs)


class gp_misc_applier(gp_applier):
    """Group Policy Miscellaneous Applier/Unapplier/Modifier
    """

    def generate_value(self, **kwargs):
        data = etree.Element('data')
        for k, v in kwargs.items():
            arg = etree.SubElement(data, k)
            arg.text = get_string(v)
        return get_string(etree.tostring(data, 'utf-8'))

    def parse_value(self, value):
        vals = {}
        try:
            data = etree.fromstring(value)
        except etree.ParseError:
            # If parsing fails, then it's an old cache value
            return {'old_val': value}
        except TypeError:
            return {}
        itr = data.iter()
        next(itr) # Skip the top element
        for item in itr:
            vals[item.tag] = item.text
        return vals


class gp_file_applier(gp_applier):
    """Group Policy File Applier/Unapplier/Modifier
    Subclass of abstract class gp_applier for monitoring policy applied
    via a file.
    """

    def __generate_value(self, value_hash, files, sep):
        data = [value_hash]
        data.extend(files)
        return sep.join(data)

    def __parse_value(self, value, sep):
        """Parse a value
        return          - A unique HASH, followed by the file list
        """
        if value is None:
            return None, []
        data = value.split(sep)
        if '/' in data[0]:
            # The first element is not a hash, but a filename. This is a
            # legacy value.
            return None, data
        else:
            return data[0], data[1:] if len(data) > 1 else []

    def unapply(self, guid, attribute, files, sep=':'):
        # If the value isn't a list of files, parse value from the log
        if type(files) != list:
            _, files = self.__parse_value(files, sep)
        for file in files:
            if os.path.exists(file):
                os.unlink(file)
        self.cache_remove_attribute(guid, attribute)

    def apply(self, guid, attribute, value_hash, applier_func, *args, sep=':'):
        """
        applier_func MUST return a list of files created by the applier.

        This applier is for policies which only apply to a single file (with
        a couple small exceptions). This applier will remove any policy applied
        by this GPO which doesn't match the new policy.
        """
        # If the policy has changed, unapply, then apply new policy
        old_val = self.cache_get_attribute_value(guid, attribute)
        # Ignore removal if this policy is applied and hasn't changed
        old_val_hash, old_val_files = self.__parse_value(old_val, sep)
        if (old_val_hash != value_hash or
                self.cache_get_apply_state() == GPOSTATE.ENFORCE) or \
                not all([os.path.exists(f) for f in old_val_files]):
            self.unapply(guid, attribute, old_val_files)
        else:
            # If policy is already applied, skip application
            return

        # Apply the policy and log the changes
        files = applier_func(*args)
        new_value = self.__generate_value(value_hash, files, sep)
        self.cache_add_attribute(guid, attribute, new_value)


""" Fetch the hostname of a writable DC """


def get_dc_hostname(creds, lp):
    net = Net(creds=creds, lp=lp)
    cldap_ret = net.finddc(domain=lp.get('realm'), flags=(nbt.NBT_SERVER_LDAP |
                                                          nbt.NBT_SERVER_DS))
    return cldap_ret.pdc_dns_name


""" Fetch a list of GUIDs for applicable GPOs """


def get_gpo(samdb, gpo_dn):
    g = gpo.GROUP_POLICY_OBJECT()
    attrs = [
        "cn",
        "displayName",
        "flags",
        "gPCFileSysPath",
        "gPCFunctionalityVersion",
        "gPCMachineExtensionNames",
        "gPCUserExtensionNames",
        "gPCWQLFilter",
        "name",
        "nTSecurityDescriptor",
        "versionNumber"
    ]
    if gpo_dn.startswith("LDAP://"):
        gpo_dn = gpo_dn.lstrip("LDAP://")

    sd_flags = (security.SECINFO_OWNER |
                security.SECINFO_GROUP |
                security.SECINFO_DACL)
    try:
        res = samdb.search(gpo_dn, ldb.SCOPE_BASE, "(objectclass=*)", attrs,
                           controls=['sd_flags:1:%d' % sd_flags])
    except Exception:
        log.error('Failed to fetch gpo object with nTSecurityDescriptor')
        raise
    if res.count != 1:
        raise ldb.LdbError(ldb.ERR_NO_SUCH_OBJECT,
                           'get_gpo: search failed')

    g.ds_path = gpo_dn
    if 'versionNumber' in res.msgs[0].keys():
        g.version = int(res.msgs[0]['versionNumber'][0])
    if 'flags' in res.msgs[0].keys():
        g.options = int(res.msgs[0]['flags'][0])
    if 'gPCFileSysPath' in res.msgs[0].keys():
        g.file_sys_path = res.msgs[0]['gPCFileSysPath'][0].decode()
    if 'displayName' in res.msgs[0].keys():
        g.display_name = res.msgs[0]['displayName'][0].decode()
    if 'name' in res.msgs[0].keys():
        g.name = res.msgs[0]['name'][0].decode()
    if 'gPCMachineExtensionNames' in res.msgs[0].keys():
        g.machine_extensions = str(res.msgs[0]['gPCMachineExtensionNames'][0])
    if 'gPCUserExtensionNames' in res.msgs[0].keys():
        g.user_extensions = str(res.msgs[0]['gPCUserExtensionNames'][0])
    if 'nTSecurityDescriptor' in res.msgs[0].keys():
        g.set_sec_desc(bytes(res.msgs[0]['nTSecurityDescriptor'][0]))
    return g

class GP_LINK:
    def __init__(self, gPLink, gPOptions):
        self.link_names = []
        self.link_opts = []
        self.gpo_parse_gplink(gPLink)
        self.gp_opts = int(gPOptions)

    def gpo_parse_gplink(self, gPLink):
        for p in gPLink.decode().split(']'):
            if not p:
                continue
            log.debug('gpo_parse_gplink: processing link')
            p = p.lstrip('[')
            link_name, link_opt = p.split(';')
            log.debug('gpo_parse_gplink: link: {}'.format(link_name))
            log.debug('gpo_parse_gplink: opt: {}'.format(link_opt))
            self.link_names.append(link_name)
            self.link_opts.append(int(link_opt))

    def num_links(self):
        if len(self.link_names) != len(self.link_opts):
            raise RuntimeError('Link names and opts mismatch')
        return len(self.link_names)

def find_samaccount(samdb, samaccountname):
    attrs = ['dn', 'userAccountControl']
    res = samdb.search(samdb.get_default_basedn(), ldb.SCOPE_SUBTREE,
                       '(sAMAccountName={})'.format(samaccountname), attrs)
    if res.count != 1:
        raise ldb.LdbError(ldb.ERR_NO_SUCH_OBJECT,
            "Failed to find samAccountName '{}'".format(samaccountname)
        )
    uac = int(res.msgs[0]['userAccountControl'][0])
    dn = res.msgs[0]['dn']
    log.info('Found dn {} for samaccountname {}'.format(dn, samaccountname))
    return uac, dn

def get_gpo_link(samdb, link_dn):
    res = samdb.search(link_dn, ldb.SCOPE_BASE,
                       '(objectclass=*)', ['gPLink', 'gPOptions'])
    if res.count != 1:
        raise ldb.LdbError(ldb.ERR_NO_SUCH_OBJECT, 'get_gpo_link: no result')
    if 'gPLink' not in res.msgs[0]:
        raise ldb.LdbError(ldb.ERR_NO_SUCH_ATTRIBUTE,
            "get_gpo_link: no 'gPLink' attribute found for '{}'".format(link_dn)
        )
    gPLink = res.msgs[0]['gPLink'][0]
    gPOptions = 0
    if 'gPOptions' in res.msgs[0]:
        gPOptions = res.msgs[0]['gPOptions'][0]
    else:
        log.debug("get_gpo_link: no 'gPOptions' attribute found")
    return GP_LINK(gPLink, gPOptions)

def add_gplink_to_gpo_list(samdb, gpo_list, forced_gpo_list, link_dn, gp_link,
                           link_type, only_add_forced_gpos, token):
    for i in range(gp_link.num_links()-1, -1, -1):
        is_forced = (gp_link.link_opts[i] & GPLINK_OPT_ENFORCE) != 0
        if gp_link.link_opts[i] & GPLINK_OPT_DISABLE:
            log.debug('skipping disabled GPO')
            continue

        if only_add_forced_gpos:
            if not is_forced:
                log.debug("skipping nonenforced GPO link "
                          "because GPOPTIONS_BLOCK_INHERITANCE "
                          "has been set")
                continue
            else:
                log.debug("adding enforced GPO link although "
                          "the GPOPTIONS_BLOCK_INHERITANCE "
                          "has been set")

        try:
            new_gpo = get_gpo(samdb, gp_link.link_names[i])
        except ldb.LdbError as e:
            (enum, estr) = e.args
            log.debug("failed to get gpo: %s" % gp_link.link_names[i])
            if enum == ldb.ERR_NO_SUCH_OBJECT:
                log.debug("skipping empty gpo: %s" % gp_link.link_names[i])
                continue
            return
        else:
            try:
                sec_desc = ndr_unpack(security.descriptor,
                                      new_gpo.get_sec_desc_buf())
                samba.security.access_check(sec_desc, token,
                                            security.SEC_STD_READ_CONTROL |
                                            security.SEC_ADS_LIST |
                                            security.SEC_ADS_READ_PROP)
            except Exception as e:
                log.debug("skipping GPO \"%s\" as object "
                          "has no access to it" % new_gpo.display_name)
                continue

            new_gpo.link = str(link_dn)
            new_gpo.link_type = link_type

            if is_forced:
                forced_gpo_list.insert(0, new_gpo)
            else:
                gpo_list.insert(0, new_gpo)

            log.debug("add_gplink_to_gpo_list: added GPLINK #%d %s "
                      "to GPO list" % (i, gp_link.link_names[i]))

def merge_with_system_token(token_1):
    sids = token_1.sids
    system_token = system_session().security_token
    sids.extend(system_token.sids)
    token_1.sids = sids
    token_1.rights_mask |= system_token.rights_mask
    token_1.privilege_mask |= system_token.privilege_mask
    # There are no claims in the system token, so it is safe not to merge the claims
    return token_1


def site_dn_for_machine(samdb, dc_hostname, lp, creds, hostname):
    # [MS-GPOL] 3.2.5.1.4 Site Search

    # The netr_DsRGetSiteName() needs to run over local rpc, however we do not
    # have the call implemented in our rpc_server.
    # What netr_DsRGetSiteName() actually does is an ldap query to get
    # the sitename, we can do the same.

    # NtVer=(NETLOGON_NT_VERSION_IP|NETLOGON_NT_VERSION_WITH_CLOSEST_SITE|
    #        NETLOGON_NT_VERSION_5EX) [0x20000014]
    expr = "(&(DnsDomain=%s.)(User=%s)(NtVer=\\14\\00\\00\\20))" % (
        samdb.domain_dns_name(),
        hostname)
    res = samdb.search(
        base='',
        scope=ldb.SCOPE_BASE,
        expression=expr,
        attrs=["Netlogon"])
    if res.count != 1:
        raise RuntimeError('site_dn_for_machine: No result')

    samlogon_response = ndr_unpack(nbt.netlogon_samlogon_response,
                                   bytes(res.msgs[0]['Netlogon'][0]))
    if not (samlogon_response.ntver & nbt.NETLOGON_NT_VERSION_5EX):
        raise RuntimeError('site_dn_for_machine: Invalid NtVer in '
                           + 'netlogon_samlogon_response')

    # We want NETLOGON_NT_VERSION_5EX out of the union!
    samlogon_response.ntver = nbt.NETLOGON_NT_VERSION_5EX
    samlogon_response_ex = samlogon_response.data

    client_site = "Default-First-Site-Name"
    if (samlogon_response_ex.client_site
            and len(samlogon_response_ex.client_site) > 1):
        client_site = samlogon_response_ex.client_site

    site_dn = samdb.get_config_basedn()
    site_dn.add_child("CN=Sites")
    site_dn.add_child("CN=%s" % (client_site))

    return site_dn



def get_gpo_list(dc_hostname, creds, lp, username):
    """Get the full list of GROUP_POLICY_OBJECTs for a given username.
    Push GPOs to gpo_list so that the traversal order of the list matches
    the order of application:
    (L)ocal (S)ite (D)omain (O)rganizational(U)nit
    For different domains and OUs: parent-to-child.
    Within same level of domains and OUs: Link order.
    Since GPOs are pushed to the front of gpo_list, GPOs have to be
    pushed in the opposite order of application (OUs first, local last,
    child-to-parent).
    Forced GPOs are appended in the end since they override all others.
    """
    gpo_list = []
    forced_gpo_list = []
    url = 'ldap://' + dc_hostname
    samdb = SamDB(url=url,
                  session_info=system_session(),
                  credentials=creds, lp=lp)
    # username is DOM\\SAM, but get_gpo_list expects SAM
    uac, dn = find_samaccount(samdb, username.split('\\')[-1])
    add_only_forced_gpos = False

    # Fetch the security token
    session_info_flags = (AUTH_SESSION_INFO_DEFAULT_GROUPS |
                          AUTH_SESSION_INFO_AUTHENTICATED)
    if url.startswith('ldap'):
        session_info_flags |= AUTH_SESSION_INFO_SIMPLE_PRIVILEGES
    session = samba.auth.user_session(samdb, lp_ctx=lp, dn=dn,
                                      session_info_flags=session_info_flags)
    gpo_list_machine = False
    if uac & UF_WORKSTATION_TRUST_ACCOUNT or uac & UF_SERVER_TRUST_ACCOUNT:
        gpo_list_machine = True
        token = merge_with_system_token(session.security_token)
    else:
        token = session.security_token

    # (O)rganizational(U)nit
    parent_dn = dn.parent()
    while True:
        if str(parent_dn) == str(samdb.get_default_basedn().parent()):
            break

        # An account can be a member of more OUs
        if parent_dn.get_component_name(0) == 'OU':
            try:
                log.debug("get_gpo_list: query OU: [%s] for GPOs" % parent_dn)
                gp_link = get_gpo_link(samdb, parent_dn)
            except ldb.LdbError as e:
                (enum, estr) = e.args
                log.debug(estr)
            else:
                add_gplink_to_gpo_list(samdb, gpo_list, forced_gpo_list,
                                       parent_dn, gp_link,
                                       gpo.GP_LINK_OU,
                                       add_only_forced_gpos, token)

                # block inheritance from now on
                if gp_link.gp_opts & GPO_BLOCK_INHERITANCE:
                    add_only_forced_gpos = True

        parent_dn = parent_dn.parent()

    # (D)omain
    parent_dn = dn.parent()
    while True:
        if str(parent_dn) == str(samdb.get_default_basedn().parent()):
            break

        # An account can just be a member of one domain
        if parent_dn.get_component_name(0) == 'DC':
            try:
                log.debug("get_gpo_list: query DC: [%s] for GPOs" % parent_dn)
                gp_link = get_gpo_link(samdb, parent_dn)
            except ldb.LdbError as e:
                (enum, estr) = e.args
                log.debug(estr)
            else:
                add_gplink_to_gpo_list(samdb, gpo_list, forced_gpo_list,
                                       parent_dn, gp_link,
                                       gpo.GP_LINK_DOMAIN,
                                       add_only_forced_gpos, token)

                # block inheritance from now on
                if gp_link.gp_opts & GPO_BLOCK_INHERITANCE:
                    add_only_forced_gpos = True

        parent_dn = parent_dn.parent()

    # (S)ite
    if gpo_list_machine:
        try:
            site_dn = site_dn_for_machine(samdb, dc_hostname, lp, creds, username)

            try:
                log.debug("get_gpo_list: query SITE: [%s] for GPOs" % site_dn)
                gp_link = get_gpo_link(samdb, site_dn)
            except ldb.LdbError as e:
                (enum, estr) = e.args
                log.debug(estr)
            else:
                add_gplink_to_gpo_list(samdb, gpo_list, forced_gpo_list,
                                       site_dn, gp_link,
                                       gpo.GP_LINK_SITE,
                                       add_only_forced_gpos, token)
        except ldb.LdbError:
            # [MS-GPOL] 3.2.5.1.4 Site Search: If the method returns
            # ERROR_NO_SITENAME, the remainder of this message MUST be skipped
            # and the protocol sequence MUST continue at GPO Search
            pass

    # (L)ocal
    gpo_list.insert(0, gpo.GROUP_POLICY_OBJECT("Local Policy",
                                               "Local Policy",
                                               gpo.GP_LINK_LOCAL))

    # Append |forced_gpo_list| at the end of |gpo_list|,
    # so that forced GPOs are applied on top of non enforced GPOs.
    return gpo_list+forced_gpo_list


def cache_gpo_dir(conn, cache, sub_dir):
    loc_sub_dir = sub_dir.upper()
    local_dir = os.path.join(cache, loc_sub_dir)
    try:
        os.makedirs(local_dir, mode=0o755)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise
    for fdata in conn.list(sub_dir):
        if fdata['attrib'] & libsmb.FILE_ATTRIBUTE_DIRECTORY:
            cache_gpo_dir(conn, cache, os.path.join(sub_dir, fdata['name']))
        else:
            local_name = fdata['name'].upper()
            f = NamedTemporaryFile(delete=False, dir=local_dir)
            fname = os.path.join(sub_dir, fdata['name']).replace('/', '\\')
            f.write(conn.loadfile(fname))
            f.close()
            os.rename(f.name, os.path.join(local_dir, local_name))


def check_safe_path(path):
    dirs = re.split('/|\\\\', path)
    if 'sysvol' in path.lower():
        ldirs = re.split('/|\\\\', path.lower())
        dirs = dirs[ldirs.index('sysvol') + 1:]
    if '..' not in dirs:
        return os.path.join(*dirs)
    raise OSError(path)


def check_refresh_gpo_list(dc_hostname, lp, creds, gpos):
    # Force signing for the connection
    saved_signing_state = creds.get_smb_signing()
    creds.set_smb_signing(SMB_SIGNING_REQUIRED)
    conn = libsmb.Conn(dc_hostname, 'sysvol', lp=lp, creds=creds)
    # Reset signing state
    creds.set_smb_signing(saved_signing_state)
    cache_path = lp.cache_path('gpo_cache')
    for gpo_obj in gpos:
        if not gpo_obj.file_sys_path:
            continue
        cache_gpo_dir(conn, cache_path, check_safe_path(gpo_obj.file_sys_path))


def get_deleted_gpos_list(gp_db, gpos):
    applied_gpos = gp_db.get_applied_guids()
    current_guids = set([p.name for p in gpos])
    deleted_gpos = [guid for guid in applied_gpos if guid not in current_guids]
    return gp_db.get_applied_settings(deleted_gpos)

def gpo_version(lp, path):
    # gpo.gpo_get_sysvol_gpt_version() reads the GPT.INI from a local file,
    # read from the gpo client cache.
    gpt_path = lp.cache_path(os.path.join('gpo_cache', path))
    return int(gpo.gpo_get_sysvol_gpt_version(gpt_path)[1])


def apply_gp(lp, creds, store, gp_extensions, username, target, force=False):
    gp_db = store.get_gplog(username)
    dc_hostname = get_dc_hostname(creds, lp)
    gpos = get_gpo_list(dc_hostname, creds, lp, username)
    del_gpos = get_deleted_gpos_list(gp_db, gpos)
    try:
        check_refresh_gpo_list(dc_hostname, lp, creds, gpos)
    except:
        log.error('Failed downloading gpt cache from \'%s\' using SMB'
                  % dc_hostname)
        return

    if force:
        changed_gpos = gpos
        gp_db.state(GPOSTATE.ENFORCE)
    else:
        changed_gpos = []
        for gpo_obj in gpos:
            if not gpo_obj.file_sys_path:
                continue
            guid = gpo_obj.name
            path = check_safe_path(gpo_obj.file_sys_path).upper()
            version = gpo_version(lp, path)
            if version != store.get_int(guid):
                log.info('GPO %s has changed' % guid)
                changed_gpos.append(gpo_obj)
        gp_db.state(GPOSTATE.APPLY)

    store.start()
    for ext in gp_extensions:
        try:
            ext = ext(lp, creds, username, store)
            if target == 'Computer':
                ext.process_group_policy(del_gpos, changed_gpos)
            else:
                drop_privileges(username, ext.process_group_policy,
                                del_gpos, changed_gpos)
        except Exception as e:
            log.error('Failed to apply extension  %s' % str(ext))
            _, _, tb = sys.exc_info()
            filename, line_number, _, _ = traceback.extract_tb(tb)[-1]
            log.error('%s:%d: %s: %s' % (filename, line_number,
                                         type(e).__name__, str(e)))
            continue
    for gpo_obj in gpos:
        if not gpo_obj.file_sys_path:
            continue
        guid = gpo_obj.name
        path = check_safe_path(gpo_obj.file_sys_path).upper()
        version = gpo_version(lp, path)
        store.store(guid, '%i' % version)
    store.commit()


def unapply_gp(lp, creds, store, gp_extensions, username, target):
    gp_db = store.get_gplog(username)
    gp_db.state(GPOSTATE.UNAPPLY)
    # Treat all applied gpos as deleted
    del_gpos = gp_db.get_applied_settings(gp_db.get_applied_guids())
    store.start()
    for ext in gp_extensions:
        try:
            ext = ext(lp, creds, username, store)
            if target == 'Computer':
                ext.process_group_policy(del_gpos, [])
            else:
                drop_privileges(username, ext.process_group_policy,
                                del_gpos, [])
        except Exception as e:
            log.error('Failed to unapply extension  %s' % str(ext))
            log.error('Message was: ' + str(e))
            continue
    store.commit()


def __rsop_vals(vals, level=4):
    if type(vals) == dict:
        ret = [' '*level + '[ %s ] = %s' % (k, __rsop_vals(v, level+2))
                for k, v in vals.items()]
        return '\n' + '\n'.join(ret)
    elif type(vals) == list:
        ret = [' '*level + '[ %s ]' % __rsop_vals(v, level+2) for v in vals]
        return '\n' + '\n'.join(ret)
    else:
        if isinstance(vals, numbers.Number):
            return ' '*(level+2) + str(vals)
        else:
            return ' '*(level+2) + get_string(vals)

def rsop(lp, creds, store, gp_extensions, username, target):
    dc_hostname = get_dc_hostname(creds, lp)
    gpos = get_gpo_list(dc_hostname, creds, lp, username)
    check_refresh_gpo_list(dc_hostname, lp, creds, gpos)

    print('Resultant Set of Policy')
    print('%s Policy\n' % target)
    term_width = shutil.get_terminal_size(fallback=(120, 50))[0]
    for gpo_obj in gpos:
        if gpo_obj.display_name.strip() == 'Local Policy':
            continue # We never apply local policy
        print('GPO: %s' % gpo_obj.display_name)
        print('='*term_width)
        for ext in gp_extensions:
            ext = ext(lp, creds, username, store)
            cse_name_m = re.findall(r"'([\w\.]+)'", str(type(ext)))
            if len(cse_name_m) > 0:
                cse_name = cse_name_m[-1].split('.')[-1]
            else:
                cse_name = ext.__module__.split('.')[-1]
            print('  CSE: %s' % cse_name)
            print('  ' + ('-'*int(term_width/2)))
            for section, settings in ext.rsop(gpo_obj).items():
                print('    Policy Type: %s' % section)
                print('    ' + ('-'*int(term_width/2)))
                print(__rsop_vals(settings).lstrip('\n'))
                print('    ' + ('-'*int(term_width/2)))
            print('  ' + ('-'*int(term_width/2)))
        print('%s\n' % ('='*term_width))


def parse_gpext_conf(smb_conf):
    from samba.samba3 import param as s3param
    lp = s3param.get_context()
    if smb_conf is not None:
        lp.load(smb_conf)
    else:
        lp.load_default()
    ext_conf = lp.state_path('gpext.conf')
    parser = ConfigParser(interpolation=None)
    parser.read(ext_conf)
    return lp, parser


def atomic_write_conf(lp, parser):
    ext_conf = lp.state_path('gpext.conf')
    with NamedTemporaryFile(mode="w+", delete=False, dir=os.path.dirname(ext_conf)) as f:
        parser.write(f)
        os.rename(f.name, ext_conf)


def check_guid(guid):
    # Check for valid guid with curly braces
    if guid[0] != '{' or guid[-1] != '}' or len(guid) != 38:
        return False
    try:
        UUID(guid, version=4)
    except ValueError:
        return False
    return True


def register_gp_extension(guid, name, path,
                          smb_conf=None, machine=True, user=True):
    # Check that the module exists
    if not os.path.exists(path):
        return False
    if not check_guid(guid):
        return False

    lp, parser = parse_gpext_conf(smb_conf)
    if guid not in parser.sections():
        parser.add_section(guid)
    parser.set(guid, 'DllName', path)
    parser.set(guid, 'ProcessGroupPolicy', name)
    parser.set(guid, 'NoMachinePolicy', "0" if machine else "1")
    parser.set(guid, 'NoUserPolicy', "0" if user else "1")

    atomic_write_conf(lp, parser)

    return True


def list_gp_extensions(smb_conf=None):
    _, parser = parse_gpext_conf(smb_conf)
    results = {}
    for guid in parser.sections():
        results[guid] = {}
        results[guid]['DllName'] = parser.get(guid, 'DllName')
        results[guid]['ProcessGroupPolicy'] = \
            parser.get(guid, 'ProcessGroupPolicy')
        results[guid]['MachinePolicy'] = \
            not int(parser.get(guid, 'NoMachinePolicy'))
        results[guid]['UserPolicy'] = not int(parser.get(guid, 'NoUserPolicy'))
    return results


def unregister_gp_extension(guid, smb_conf=None):
    if not check_guid(guid):
        return False

    lp, parser = parse_gpext_conf(smb_conf)
    if guid in parser.sections():
        parser.remove_section(guid)

    atomic_write_conf(lp, parser)

    return True


def set_privileges(username, uid, gid):
    """
    Set current process privileges
    """

    os.setegid(gid)
    os.seteuid(uid)


def drop_privileges(username, func, *args):
    """
    Run supplied function with privileges for specified username.
    """
    current_uid = os.getuid()

    if not current_uid == 0:
        raise Exception('Not enough permissions to drop privileges')

    user_uid = pwd.getpwnam(username).pw_uid
    user_gid = pwd.getpwnam(username).pw_gid

    # Drop privileges
    set_privileges(username, user_uid, user_gid)

    # We need to catch exception in order to be able to restore
    # privileges later in this function
    out = None
    exc = None
    try:
        out = func(*args)
    except Exception as e:
        exc = e

    # Restore privileges
    set_privileges('root', current_uid, 0)

    if exc:
        raise exc

    return out

def expand_pref_variables(text, gpt_path, lp, username=None):
    utc_dt = datetime.utcnow()
    dt = datetime.now()
    cache_path = lp.cache_path(os.path.join('gpo_cache'))
    # These are all the possible preference variables that MS supports. The
    # variables set to 'None' here are currently unsupported by Samba, and will
    # prevent the individual policy from applying.
    variables = { 'AppDataDir': os.path.expanduser('~/.config'),
                  'BinaryComputerSid': None,
                  'BinaryUserSid': None,
                  'CommonAppdataDir': None,
                  'CommonDesktopDir': None,
                  'CommonFavoritesDir': None,
                  'CommonProgramsDir': None,
                  'CommonStartUpDir': None,
                  'ComputerName': lp.get('netbios name'),
                  'CurrentProccessId': None,
                  'CurrentThreadId': None,
                  'DateTime': utc_dt.strftime('%Y-%m-%d %H:%M:%S UTC'),
                  'DateTimeEx': str(utc_dt),
                  'DesktopDir': os.path.expanduser('~/Desktop'),
                  'DomainName': lp.get('realm'),
                  'FavoritesDir': None,
                  'GphPath': None,
                  'GptPath': os.path.join(cache_path,
                                          check_safe_path(gpt_path).upper()),
                  'GroupPolicyVersion': None,
                  'LastDriveMapped': None,
                  'LastError': None,
                  'LastErrorText': None,
                  'LdapComputerSid': None,
                  'LdapUserSid': None,
                  'LocalTime': dt.strftime('%H:%M:%S'),
                  'LocalTimeEx': dt.strftime('%H:%M:%S.%f'),
                  'LogonDomain': lp.get('realm'),
                  'LogonServer': None,
                  'LogonUser': username,
                  'LogonUserSid': None,
                  'MacAddress': None,
                  'NetPlacesDir': None,
                  'OsVersion': None,
                  'ProgramFilesDir': None,
                  'ProgramsDir': None,
                  'RecentDocumentsDir': None,
                  'ResultCode': None,
                  'ResultText': None,
                  'ReversedComputerSid': None,
                  'ReversedUserSid': None,
                  'SendToDir': None,
                  'StartMenuDir': None,
                  'StartUpDir': None,
                  'SystemDir': None,
                  'SystemDrive': '/',
                  'TempDir': '/tmp',
                  'TimeStamp': str(datetime.timestamp(dt)),
                  'TraceFile': None,
                  'WindowsDir': None
    }
    for exp_var, val in variables.items():
        exp_var_fmt = '%%%s%%' % exp_var
        if exp_var_fmt in text:
            if val is None:
                raise NameError('Expansion variable %s is undefined' % exp_var)
            text = text.replace(exp_var_fmt, val)
    return text
